//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// james@akrobiz.com
//
//############################################################################
#include "ez_uail_gui.hpp"

#define CHARS_IN_LINE 80
    
//############################################################################
void clear_string(char formatted_string[80])
{
    int i;
    for(i = 0; i < CHARS_IN_LINE; i++)
        formatted_string[i] = 0x00;
    return;
}

//############################################################################
int main(int argc, char *argv[])
{
    bool     running              = true ,
             show_help            = true ,
             in_sub_menu          = false,
             ui_visuals_flip_flop = false,
             show_frame_stats     = false,
             show_frame_set_stats = false,
             teminate_requested   = false;
    u_int    xres                 = 640  ,
             yres                 = 480  ,
             time_zero                   ;             
    int      temp_int                    ;
    float    temp_float                  ;
    string   file_name                   ,
             ui_message                  ;
    char     formatted_string[80]        ;
    ez_color bg_was                      ;

    if(argc >= 2)
    {
        xres = atoi(argv[1]);
        yres = atoi(argv[2]);
    }

    ez_uail_gui ail(xres, yres); // this is it!

    ail.from_file("in.ild");
    if(ail.error)
        ui_message = "NO DATA!";
    else
        ui_message = "in.ild";

    clear_string(formatted_string);
    ail.display(ui_message);
    ail.display_program_menu();

    while(running)
    {                         
        do
        {                                      
            SDL_WaitEvent(&event);
        } while((event.type != SDL_KEYDOWN));

        switch(event.key.keysym.unicode)
        {
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                teminate_requested = true;
                break;
            //----------------------------------------------------------------




            //----------------------------------------------------------------
            case SDLK_RETURN:
                show_help = !show_help;
                if(!show_help)
                    ui_message = "[Enter] help";
                break;
            //----------------------------------------------------------------




            //----------------------------------------------------------------
            default:
                switch(event.key.keysym.sym)
                {
                    //--------------------------------------------------------
                    case SDLK_UP: // display next palette
                        ail.target_palette_index++;
                        if(ail.target_palette_index >= ail.number_of_palettes())
                            ail.target_palette_index = 0;
                        ui_message = "+ target palette";
                        break;
                    //--------------------------------------------------------
                    case SDLK_DOWN: // display previous palette
                        ail.target_palette_index--;
                        if(ail.target_palette_index < 0)
                            ail.target_palette_index = ail.number_of_palettes() - 1;
                        ui_message = "- target palette";
                        break;                    
                    //--------------------------------------------------------
                    case SDLK_RIGHT: // display previous frame
                        if(    ail.loop_animation
                            || (ail.frame_index < ail.number_of_frames() - 1)
                          )
                            ++ail;
                        ui_message = "+ frame";
                        break;
                    //--------------------------------------------------------
                    case SDLK_LEFT: // display next frame
                        if(    ail.loop_animation
                            || (ail.frame_index > 0)
                          )
                            --ail;
                        ui_message = "- frame";
                        break;
                }
                break;
            //----------------------------------------------------------------





            //----------------------------------------------------------------
            case SDLK_BACKSPACE:
                ail.delete_current_frame();
                ui_message = "deleted";
                break;
            //---------------------------------------------------------------- 





            //----------------------------------------------------------------
            case '`': // start animation forward
                SDL_Delay(250);
                SDL_PollEvent(&event);
                while(    (event.type != SDL_KEYDOWN)
                       && (    ail.loop_animation
                            || (ail.frame_index < ail.number_of_frames() - 1)
                          )
                     )
                {
                    time_zero = SDL_GetTicks();
                    ++ail;
                    ail.display("+ show");
                    if(ail.approximate_frame_rate)
                    {
                        do{;}
                        while(   SDL_GetTicks() - time_zero
                               < (u_int)((1.0 / ail.frames_per_second) * 1000.0)
                             );
                    }
                    SDL_PollEvent(&event);
                }
                break;
            //----------------------------------------------------------------
            case '~': // start animation reverse
                SDL_Delay(250);
                SDL_PollEvent(&event);
                while(    (event.type != SDL_KEYDOWN)
                       && (    ail.loop_animation
                            || (ail.frame_index > 0)
                          )
                     )
                {
                    time_zero = SDL_GetTicks();
                    --ail;
                    ail.display("- show");
                    if(ail.approximate_frame_rate)
                    {
                        do{;}
                        while(   SDL_GetTicks() - time_zero
                               < (u_int)((1.0 / ail.frames_per_second) * 1000.0)
                             );
                    }
                    SDL_PollEvent(&event);
                }
                break;
            //----------------------------------------------------------------




            //----------------------------------------------------------------
            case '-': // start rotation formard
                SDL_Delay(250);
                SDL_PollEvent(&event);
                while(    (event.type != SDL_KEYDOWN)
                       && (    ail.loop_animation
                            || (ail.frame_index < ail.number_of_frames() - 1)
                          )
                     )
                {
                    time_zero = SDL_GetTicks();
                    ail.view_angle += one_degree;
                    ail.display("+ orbit");
                    if(ail.approximate_frame_rate)
                    {
                        do{;}
                        while(   SDL_GetTicks() - time_zero
                               < (u_int)((1.0 / ail.frames_per_second) * 1000.0)
                             );
                    }
                    SDL_PollEvent(&event);
                }
                break;
            //----------------------------------------------------------------
            case '_': // start rotation reverse
                SDL_Delay(250);
                SDL_PollEvent(&event);
                while(    (event.type != SDL_KEYDOWN)
                       && (    ail.loop_animation
                            || (ail.frame_index > 0)
                          )
                     )
                {
                    time_zero = SDL_GetTicks();
                    ail.view_angle -= one_degree;
                    ail.display("- orbit");
                    if(ail.approximate_frame_rate)
                    {
                        do{;}
                        while(   SDL_GetTicks() - time_zero
                               < (u_int)((1.0 / ail.frames_per_second) * 1000.0)
                             );
                    }
                    SDL_PollEvent(&event);
                }
                break;
            //----------------------------------------------------------------





            //----------------------------------------------------------------
            case '=': // start rotation formard
                SDL_Delay(250);
                SDL_PollEvent(&event);
                while(    (event.type != SDL_KEYDOWN)
                       && (    ail.loop_animation
                            || (ail.frame_index < ail.number_of_frames() - 1)
                          )
                     )
                {
                    time_zero = SDL_GetTicks();
                    ++ail;
                    ail.view_angle += one_degree;
                    ail.display("+ show & orbit");
                    if(ail.approximate_frame_rate)
                    {
                        do{;}
                        while(   SDL_GetTicks() - time_zero
                               < (u_int)((1.0 / ail.frames_per_second) * 1000.0)
                             );
                    }
                    SDL_PollEvent(&event);
                }
                break;
            //----------------------------------------------------------------
            case '+': // start rotation reverse
                SDL_Delay(250);
                SDL_PollEvent(&event);
                while(    (event.type != SDL_KEYDOWN)
                       && (    ail.loop_animation
                            || (ail.frame_index > 0)
                          )
                     )
                {
                    time_zero = SDL_GetTicks();
                    --ail;
                    ail.view_angle -= one_degree;
                    ail.display("- show & orbit");
                    if(ail.approximate_frame_rate)
                    {
                        do{;}
                        while(   SDL_GetTicks() - time_zero
                               < (u_int)((1.0 / ail.frames_per_second) * 1000.0)
                             );
                    }
                    SDL_PollEvent(&event);
                }
                break;
            //----------------------------------------------------------------




            //----------------------------------------------------------------
            case '/': // show frame set stats
                show_frame_set_stats = !show_frame_set_stats;
                break;                    
            //----------------------------------------------------------------
            case '?': // show frame stats
                show_frame_stats = !show_frame_stats;
                break;                    
            //----------------------------------------------------------------





            //----------------------------------------------------------------
            case ';': // toggle bitmap background
                ail.show_background_bitmap = !ail.show_background_bitmap;
                break;                    
            //----------------------------------------------------------------





            //----------------------------------------------------------------
            case '1':
                ail.view_angle.x += ail.rotate_radians_per_key;
                ui_message = "+ orbit X";
                break;                    
            //----------------------------------------------------------------
            case '!':
                ail.view_angle.x -= ail.rotate_radians_per_key;
                ui_message = "- orbit X";
                break;                    
            //----------------------------------------------------------------
            case '2':
                ail.view_angle.y += ail.rotate_radians_per_key;
                ui_message = "+ orbit Y";
                break;                    
            //----------------------------------------------------------------
            case '@':
                ail.view_angle.y -= ail.rotate_radians_per_key;
                ui_message = "- orbit Y";
                break;                    
            //----------------------------------------------------------------
            case '3':
                ail.view_angle.z += ail.rotate_radians_per_key;
                ui_message = "+ orbit Z";
                break;                    
            //----------------------------------------------------------------
            case '#':
                ail.view_angle.z -= ail.rotate_radians_per_key;
                ui_message = "- orbit Z";
                break;                    
            //----------------------------------------------------------------
            case '4':
                ail.view_angle += ail.rotate_radians_per_key;
                ui_message = "+ orbit XYZ";
                break;                    
            //----------------------------------------------------------------
            case '$':
                ail.view_angle -= ail.rotate_radians_per_key;
                ui_message = "- orbit XYZ";
                break;                    
            //----------------------------------------------------------------
            case '5':
                ail.view_angle   = 0;
                ui_message = "front";
                break;                    
            //----------------------------------------------------------------
            case '%':
                ail.view_angle.x = 0;
                ail.view_angle.y = pi;
                ail.view_angle.z = 0;
                ui_message = "back";
                break;                    
            //----------------------------------------------------------------
            case '6':
                ail.view_angle.x = 0;
                ail.view_angle.y = half_pi;
                ail.view_angle.z = 0;
                ui_message = "left side";
                break;                    
            //----------------------------------------------------------------
            case '^':
                ail.view_angle.x = 0;
                ail.view_angle.y = -half_pi;
                ail.view_angle.z = 0;
                ui_message = "right side";
                break;                    
            //----------------------------------------------------------------
            case '7':
                ail.view_angle.x = half_pi;
                ail.view_angle.y = 0;
                ail.view_angle.z = 0;
                ui_message = "top";
                break;                    
            //----------------------------------------------------------------
            case '&':
                ail.view_angle.x = -half_pi;
                ail.view_angle.y = 0;
                ail.view_angle.z = 0;
                ui_message = "bottom";
                break;                    
            //----------------------------------------------------------------
            case '8':
                ail.view_angle += ail.rotate_radians_per_key;
                if(    ail.loop_animation
                    || (ail.frame_index < ail.number_of_frames() - 1)
                  )
                    ++ail;
                ui_message = "+ frame & orbit";
                break;
            //----------------------------------------------------------------
            case '*':
                ail.view_angle -= ail.rotate_radians_per_key;
                if(    ail.loop_animation
                    || (ail.frame_index > 0)
                  )
                    --ail;
                ui_message = "- frame & orbit";
                break;
            //----------------------------------------------------------------
            
            
            
                
            //----------------------------------------------------------------
            case '9': // go to first frame
                ail.frame(0);
                ui_message = "first frame";                            
                break;
            //----------------------------------------------------------------
            case '0': // go to last frame
                ail.frame(ail.number_of_frames() - 1);
                ui_message = "last frame";
                break;
            //----------------------------------------------------------------
            case '(': // set, unset start frame select
                ail.toggle_frame_select_start();
                if(ail.frame_select_start != -1)
                {
                    sprintf(formatted_string, "start = %d", ail.frame_select_start);
                    ui_message = formatted_string;
                    clear_string(formatted_string);
                }
                else
                    ui_message = "start unset";
                break;
            //----------------------------------------------------------------
            case ')': // set, unset end frame select
                ail.toggle_frame_select_end();
                if(ail.frame_select_end != -1)
                {
                    sprintf(formatted_string, "end = %d", ail.frame_select_end);
                    ui_message = formatted_string;
                    clear_string(formatted_string);                        
                }
                else
                    ui_message = "end unset";
                break;
            //----------------------------------------------------------------
            case ' ': // [Space] select, deselect current frame
                ail.current_frame().is_selected = !(ail.current_frame().is_selected);
                if(ail.current_frame().is_selected)
                    ui_message = "select frame";
                else
                    ui_message = "deselect frame";
                break;
            //----------------------------------------------------------------
            
            
            
            
            
            //----------------------------------------------------------------                    
            case 'q':
                ail.view_offset.x += ail.move_points_per_key;
                ui_message = "+ pan X";
                break;                    
            //----------------------------------------------------------------
            case 'Q':
                ail.view_offset.x -= ail.move_points_per_key;
                ui_message = "- pan X";
                break;                    
            //----------------------------------------------------------------
            case 'w':
                ail.view_offset.y += ail.move_points_per_key;
                ui_message = "+ pan Y";
                break;                    
            //----------------------------------------------------------------
            case 'W':
                ail.view_offset.y -= ail.move_points_per_key;
                ui_message = "- pan Y";
                break;                    
            //----------------------------------------------------------------
            case 'e':
                ail.view_offset += ail.move_points_per_key;
                ui_message = "+ pan XY";
                break;                    
            //----------------------------------------------------------------
            case 'E':
                ail.view_offset -= ail.move_points_per_key;
                ui_message = "- pan XY";
                break;                    
            //----------------------------------------------------------------
            case 'r':
            case 'R':
                ail.view_offset = 0.0;
                ui_message = "on center";
                break;
            //----------------------------------------------------------------
            case 'a':
                ail.view_scale.x *= 1.01;
                ui_message = "+ zoom X";
                break;                    
            //----------------------------------------------------------------
            case 'A':
                ail.view_scale.x *= 0.99;
                ui_message = "- zoom X";
                break;                    
            //----------------------------------------------------------------
            case 's':
                ail.view_scale.y *= 1.01;
                ui_message = "+ zoom Y";
                break;                    
            //----------------------------------------------------------------
            case 'S':
                ail.view_scale.y *= 0.99;
                ui_message = "- zoom Y";
                break;                    
            //----------------------------------------------------------------
            case 'd':
                ail.view_scale *= 1.01;
                ui_message = "+ zoom XY";
                break;                    
            //----------------------------------------------------------------
            case 'D':
                ail.view_scale *= 0.99;
                ui_message = "- zoom XY";
                break;                    
            //----------------------------------------------------------------
            case 'f':
            case 'F':
                ail.view_scale = 1.0;
                ui_message = "no zoom";
                break;
            //----------------------------------------------------------------                                        





            //----------------------------------------------------------------
            case SDLK_TAB:
                in_sub_menu = true;
                ui_message  = "program settings";                
                ail.display(ui_message);
                ail.display_settings_menu();                        
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "exit settings";
                                break;                                                    
                            //------------------------------------------------
                            case '1':
                                temp_float = (ail.display_prompt_and_echo_float("enter new insignificant angle in degrees") * one_degree);
                                if(ail.process_exception)
                                    break;                                
                                ail.zero_angle_tolerance = temp_float;
                                break;
                            //------------------------------------------------
                            case '2':
                                temp_float = ail.display_prompt_and_echo_float("enter insignificant distance in points");
                                if(ail.process_exception)
                                    break;                                
                                ail.insignificant_distance = temp_float;
                                break;
                            //------------------------------------------------
                            case '3':
                                temp_float = ail.display_prompt_and_echo_float("enter new maximum lit vector length in points");
                                if(ail.process_exception)
                                    break;
                                ail.lit_distance_delta_max = temp_float;
                                break;
                            //------------------------------------------------
                            case '4':
                                temp_float = ail.display_prompt_and_echo_float("enter new maximum blank vector length in points");
                                if(ail.process_exception)
                                    break;
                                ail.blank_distance_delta_max = temp_float;
                                break;
                            //------------------------------------------------
                            case '5':
                                temp_float = ail.display_prompt_and_echo_float("enter dwell in microseconds for 180 degree angle");
                                if(ail.process_exception)
                                    break;             
                                ail.longest_dwell_microsec = temp_float;
                                break;
                            //------------------------------------------------
                            case '6':
                                temp_int = ail.display_prompt_and_echo_int("enter new lit dwell overhang");
                                if(ail.process_exception)
                                    break;                                
                                ail.lit_dwell_overhang = temp_int;
                                break;
                            //------------------------------------------------
                            case '7':
                                temp_float = ail.display_prompt_and_echo_float("rendered arc segment angle in degrees");
                                if(ail.process_exception)
                                    break;                                
                                ail.rendered_arc_angle = temp_float;
                                break;
                            //------------------------------------------------
                            case '8':
                                temp_float = ail.display_prompt_and_echo_float("rotate degrees per key hit");
                                if(ail.process_exception)
                                    break;                                
                                ail.rotate_radians_per_key = temp_float * one_degree;
                                break;
                            //------------------------------------------------
                            case '9':
                                temp_int = ail.display_prompt_and_echo_int("enter new move points per key");
                                if(ail.process_exception)
                                    break;                                
                                ail.move_points_per_key = temp_int;
                                break;
                            //------------------------------------------------
                            case '0':
                                temp_float = ail.display_prompt_and_echo_float("enter new recolor palette span factor");
                                if(ail.process_exception)
                                    break;                                
                                ail.recolor_span_factor = temp_float;
                                break;
                            //------------------------------------------------
                            case 'a':
                                temp_float = ail.display_prompt_and_echo_float("enter new frames per second");
                                if(ail.process_exception)
                                    break;                                
                                ail.frames_per_second = temp_float;
                                break;
                            //------------------------------------------------
                            case 'b':
                                temp_float = ail.display_prompt_and_echo_float("enter new wave frame still output durration in seconds");
                                if(ail.process_exception)
                                    break;                                
                                ail.still_frame_duration = temp_float;
                                break;
                            //------------------------------------------------
                            case 'c':
                                temp_int = ail.display_prompt_and_echo_int("enter new output wave sample rate");
                                if(ail.process_exception)
                                    break;                                
                                ail.wave_sample_rate = temp_int;
                                break;
                            //------------------------------------------------
                            case 'd':
                                temp_int = ail.display_prompt_and_echo_int("output bitmap size in pixels square");
                                if(ail.process_exception)
                                    break;                                
                                ail.output_bmp_size = temp_int;
                                break;
                            //------------------------------------------------
                            case 'e':
                                temp_int = ail.display_prompt_and_echo_int("display fonst size factor");
                                if(ail.process_exception)
                                    break;                                
                                ail.font_size_factor = temp_int;
                                break;
                            //------------------------------------------------
                            case 'f':
                                temp_int = ail.display_prompt_and_echo_int("display background color red");
                                if(ail.process_exception)
                                    break;
                                ail.bg_color.r = temp_int;
                                ail.recolor_background();
                                break;
                            //------------------------------------------------
                            case 'g':
                                temp_int = ail.display_prompt_and_echo_int("display background color green");
                                if(ail.process_exception)
                                    break;
                                ail.bg_color.g = temp_int;
                                ail.recolor_background();
                                break;
                            //------------------------------------------------
                            case 'h':
                                temp_int = ail.display_prompt_and_echo_int("display background color blue");
                                if(ail.process_exception)
                                    break;
                                ail.bg_color.b = temp_int;
                                ail.recolor_background();
                                break;
                            //------------------------------------------------
                            case 'i':
                                temp_int = ail.display_prompt_and_echo_int("display background gray scale");
                                if(ail.process_exception)
                                    break;
                                ail.bg_color = (u_char)temp_int;
                                ail.recolor_background();
                                break;
                            //------------------------------------------------
                            case 'j':
                                temp_int = ail.display_prompt_and_echo_int("black level for palette reduction");
                                if(ail.process_exception)
                                    break;
                                ail.black_threshold = (u_char)temp_int;
                                break;
                            //------------------------------------------------
                        }
                        ail.display(ui_message);
                        ail.display_settings_menu();                        
                    }
                }
                break;
            //----------------------------------------------------------------                                        





            //----------------------------------------------------------------
            case 'u': // UI visuals
            case 'U':
                in_sub_menu = true;
                ui_message = "UI visuals";
                ail.display(ui_message);
                ail.display_visuals_menu();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "exit UI visuals";
                                break;
                            //------------------------------------------------
                            case '1':
                                ail.show_points          = !ail.show_points;
                                if(ail.show_points)
                                    ui_message = "show vertices";
                                else
                                    ui_message = "hide vertices";
                                break;
                            //------------------------------------------------
                            case '2':
                                ail.show_blanking        = !ail.show_blanking;
                                if(ail.show_blanking)
                                    ui_message = "show blanking";
                                else
                                    ui_message = "hide blanking";
                                break;
                            //------------------------------------------------
                            case '3':
                                ail.show_palettes        = !ail.show_palettes;
                                if(ail.show_palettes)
                                    ui_message = "show palettes";
                                else
                                    ui_message = "hide palettes";
                                break;
                            //------------------------------------------------
                            case '4':
                                ail.show_stats           = !ail.show_stats;
                                if(ail.show_stats)
                                    ui_message = "show stats";
                                else
                                    ui_message = "hide stats";
                                break;
                            //------------------------------------------------
                            case '5':
                                ail.show_fixed_bounds    = !ail.show_fixed_bounds;
                                if(ail.show_fixed_bounds)
                                    ui_message = "show fixed bounds";
                                else
                                    ui_message = "hide fixed bounds";
                                break;
                            //------------------------------------------------
                            case '6':
                                ail.show_fixed_origin    = !ail.show_fixed_origin;
                                if(ail.show_fixed_origin)
                                    ui_message = "show fixed origin";
                                else
                                    ui_message = "hide fixed origin";
                                break;
                            //------------------------------------------------
                            case '7':
                                ail.show_floating_axis   = !ail.show_floating_axis;
                                if(ail.show_floating_axis)
                                    ui_message = "show floating axis";
                                else
                                    ui_message = "hide floating axis";
                                break;
                            //------------------------------------------------
                            case '8':
                                ail.show_floating_bounds = !ail.show_floating_bounds;
                                if(ail.show_floating_bounds)
                                    ui_message = "show floating bounds";
                                else
                                    ui_message = "hide floating bounds";
                                break;
                            //------------------------------------------------
                            case '9':
                                ail.show_floating_h_bounds = !ail.show_floating_h_bounds;
                                if(ail.show_floating_h_bounds)
                                    ui_message = "show hyper bounds";
                                else
                                    ui_message = "hide hyper bounds";
                                break;
                            //------------------------------------------------
                            case '0':
                                ail.show_ui_message = !ail.show_ui_message;
                                if(ail.show_ui_message)
                                    ui_message = "show UI message"; 
                                else
                                    ui_message = "hide UI message";
                                break;
                            //------------------------------------------------
                            case 'a':
                                ui_visuals_flip_flop       = !ui_visuals_flip_flop;
                                ail.show_points            =  ui_visuals_flip_flop;
                                ail.show_blanking          =  ui_visuals_flip_flop;
                                ail.show_palettes          =  ui_visuals_flip_flop;
                                ail.show_stats             =  ui_visuals_flip_flop;
                                ail.show_fixed_bounds      =  ui_visuals_flip_flop;
                                ail.show_fixed_origin      =  ui_visuals_flip_flop;
                                ail.show_floating_axis     =  ui_visuals_flip_flop;
                                ail.show_floating_bounds   =  ui_visuals_flip_flop;
                                ail.show_floating_h_bounds =  ui_visuals_flip_flop;
                                ail.show_ui_message        =  ui_visuals_flip_flop;
                                if(ui_visuals_flip_flop)
                                    ui_message = "show visuals";
                                else
                                    ui_message = "hide visuals";
                                break;
                            //------------------------------------------------
                            case 'b':
                                ail.show_effects_generation = !ail.show_effects_generation;
                                if(ail.show_effects_generation)
                                    ui_message = "show live rendering";
                                else
                                    ui_message = "hide live rendering";
                                break;
                            //------------------------------------------------
                            case 'c':
                                ail.auto_flatten_z = !ail.auto_flatten_z;
                                if(ail.auto_flatten_z)
                                    ui_message = "flatten Z for wave";
                                else
                                    ui_message = "keep Z for wave";
                                break;
                            //------------------------------------------------
                            case 'd':
                                ail.expensive_minimize = !ail.expensive_minimize;
                                if(ail.expensive_minimize)
                                    ui_message = "+ equal segment remove";
                                else
                                    ui_message = "- equal segment remove";
                                break;
                            //------------------------------------------------
                            case 'e':
                                ail.invert_wave_output = !ail.invert_wave_output;
                                if(ail.invert_wave_output)
                                    ui_message = "inverted wave output";
                                else
                                    ui_message = "positive wave output";
                                break;
                            //------------------------------------------------
                            case 'f':
                                ail.auto_optimize = !ail.auto_optimize;
                                if(ail.auto_optimize)
                                    ui_message = "optimize file output";
                                else
                                    ui_message = "raw file output";
                                break;
                            //------------------------------------------------
                            case 'g':
                                ail.loop_animation = !ail.loop_animation;
                                if(ail.loop_animation)
                                    ui_message = "loop animation";
                                else
                                    ui_message = "stop animatiion at ends";
                                break;
                            //------------------------------------------------
                            case 'h':
                                ail.approximate_frame_rate = !ail.approximate_frame_rate;
                                if(ail.approximate_frame_rate)
                                    ui_message = "aproximate frame rate";
                                else
                                    ui_message = "as fast as possible";
                                break;
                            //------------------------------------------------
                            case 'i':
                                ail.auto_scale_dxf = !ail.auto_scale_dxf;
                                if(ail.auto_scale_dxf)
                                    ui_message = "scale dxf import";
                                else                                
                                    ui_message = "no scale dxf import";
                                break;
                            //------------------------------------------------
                            case 'j':
                                ail.maintain_dxf_origin = !ail.maintain_dxf_origin;
                                if(ail.maintain_dxf_origin)
                                    ui_message = "maintain dxf origin";
                                else
                                    ui_message = "ignore dxf origin";
                                break;
                            //------------------------------------------------
                            case 'k':
                                ail.black_dwell_points = !ail.black_dwell_points;
                                if(ail.black_dwell_points)
                                    ui_message = "black dwell";
                                else
                                    ui_message = "lit dwell";
                                break;
                            //------------------------------------------------
                            case 'l':
                                ail.no_black_match = !ail.no_black_match;
                                if(ail.no_black_match)
                                    ui_message = "no black match";
                                else
                                    ui_message = "black match palette";
                                break;
                            //------------------------------------------------
                            case 'm':
                                ail.bad_file = !ail.bad_file;
                                if(ail.bad_file)
                                    ui_message = "format 4,5";
                                else
                                    ui_message = "format 0,1,2,(!3)";
                                break;
                            //------------------------------------------------
                            case ';': // toggle bitmap background
                                ail.show_background_bitmap = !ail.show_background_bitmap;
                                break;                    
                            //------------------------------------------------
                        }
                    }
                    ail.display(ui_message);
                    ail.display_visuals_menu();
                }
                break;
            //----------------------------------------------------------------                                        






            //----------------------------------------------------------------
            case 'v': // apply view to frame(s) data
            case 'V':
            {
                ail.simplify_view_angle();
                in_sub_menu = true;
                ui_message = "view settings";
                ail.display(ui_message);
                ail.display_veiw_menu();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "exit view";
                                break;
                            //------------------------------------------------
                            default:
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------
                                    case SDLK_UP: // display next palette
                                        ail.target_palette_index++;
                                        if(ail.target_palette_index >= ail.number_of_palettes())
                                            ail.target_palette_index = 0;
                                        ui_message = "+ target palette";
                                        break;
                                    //----------------------------------------
                                    case SDLK_DOWN: // display previous palette
                                        ail.target_palette_index--;
                                        if(ail.target_palette_index < 0)
                                            ail.target_palette_index = ail.number_of_palettes() - 1;
                                        ui_message = "- target palette";
                                        break;                    
                                    //----------------------------------------
                                    case SDLK_RIGHT: // display previous frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index < ail.number_of_frames() - 1)
                                          )
                                            ++ail;
                                        ui_message = "+ frame";
                                        break;
                                    //----------------------------------------
                                    case SDLK_LEFT: // display next frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index > 0)
                                          )
                                            --ail;
                                        ui_message = "- frame";
                                        break;
                                }
                                break;
                            //------------------------------------------------
                            case '1':
                                ail.view_angle.x += ail.rotate_radians_per_key;
                                ui_message = "+ orbit X";
                                break;                    
                            //------------------------------------------------
                            case '!':
                                ail.view_angle.x -= ail.rotate_radians_per_key;
                                ui_message = "- orbit X";
                                break;                    
                            //------------------------------------------------
                            case '2':
                                ail.view_angle.y += ail.rotate_radians_per_key;
                                ui_message = "+ orbit Y";
                                break;                    
                            //------------------------------------------------
                            case '@':
                                ail.view_angle.y -= ail.rotate_radians_per_key;
                                ui_message = "- orbit Y";
                                break;                    
                            //------------------------------------------------
                            case '3':
                                ail.view_angle.z += ail.rotate_radians_per_key;
                                ui_message = "+ orbit Z";
                                break;                    
                            //------------------------------------------------
                            case '#':
                                ail.view_angle.z -= ail.rotate_radians_per_key;
                                ui_message = "- orbit Z";
                                break;                    
                            //------------------------------------------------
                            case '4':
                                ail.view_angle += ail.rotate_radians_per_key;
                                ui_message = "+ orbit XYZ";
                                break;                    
                            //------------------------------------------------
                            case '$':
                                ail.view_angle -= ail.rotate_radians_per_key;
                                ui_message = "- orbit XYZ";
                                break;                    
                            //------------------------------------------------
                            case '5':
                                ail.view_angle   = 0;
                                ui_message = "front";
                                break;                    
                            //------------------------------------------------
                            case '%':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = pi;
                                ail.view_angle.z = 0;
                                ui_message = "back";
                                break;                    
                            //------------------------------------------------
                            case '6':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = half_pi;
                                ail.view_angle.z = 0;
                                ui_message = "left side";
                                break;                    
                            //------------------------------------------------
                            case '^':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = -half_pi;
                                ail.view_angle.z = 0;
                                ui_message = "right side";
                                break;                    
                            //------------------------------------------------
                            case '7':
                                ail.view_angle.x = half_pi;
                                ail.view_angle.y = 0;
                                ail.view_angle.z = 0;
                                ui_message = "top";
                                break;                    
                            //------------------------------------------------
                            case '&':
                                ail.view_angle.x = -half_pi;
                                ail.view_angle.y = 0;
                                ail.view_angle.z = 0;
                                ui_message = "bottom";
                                break;                    
                            //------------------------------------------------
                            case '8':
                                ail.view_angle += ail.rotate_radians_per_key;
                                if(    ail.loop_animation
                                    || (ail.frame_index < ail.number_of_frames() - 1)
                                  )
                                    ++ail;
                                ui_message = "+ frame & orbit";
                                break;
                            //------------------------------------------------
                            case '*':
                                ail.view_angle -= ail.rotate_radians_per_key;
                                if(    ail.loop_animation
                                    || (ail.frame_index > 0)
                                  )
                                    --ail;
                                ui_message = "- frame & orbit";
                                break;
                            //------------------------------------------------
                            case '9': // go to first frame
                                ail.frame(0);
                                ui_message = "first frame";                            
                                break;
                            //------------------------------------------------
                            case '0': // go to last frame
                                ail.frame(ail.number_of_frames() - 1);
                                ui_message = "last frame";
                                break;
                            //------------------------------------------------
                            case 'q':
                                ail.view_offset.x += ail.move_points_per_key;
                                ui_message = "+ pan X";
                                break;                    
                            //------------------------------------------------
                            case 'Q':
                                ail.view_offset.x -= ail.move_points_per_key;
                                ui_message = "- pan X";
                                break;                    
                            //------------------------------------------------
                            case 'w':
                                ail.view_offset.y += ail.move_points_per_key;
                                ui_message = "+ pan Y";
                                break;                    
                            //------------------------------------------------
                            case 'W':
                                ail.view_offset.y -= ail.move_points_per_key;
                                ui_message = "- pan Y";
                                break;                    
                            //------------------------------------------------
                            case 'e':
                                ail.view_offset += ail.move_points_per_key;
                                ui_message = "+ pan XY";
                                break;                    
                            //------------------------------------------------
                            case 'E':
                                ail.view_offset -= ail.move_points_per_key;
                                ui_message = "- pan XY";
                                break;                    
                            //------------------------------------------------
                            case 'r':
                            case 'R':
                                ail.view_offset = 0.0;
                                ui_message = "on center";
                                break;
                            //------------------------------------------------
                            case 'a':
                                ail.view_scale.x *= 1.01;
                                ui_message = "+ zoom X";
                                break;                    
                            //------------------------------------------------
                            case 'A':
                                ail.view_scale.x *= 0.99;
                                ui_message = "- zoom X";
                                break;                    
                            //------------------------------------------------
                            case 's':
                                ail.view_scale.y *= 1.01;
                                ui_message = "+ zoom Y";
                                break;                    
                            //------------------------------------------------
                            case 'S':
                                ail.view_scale.y *= 0.99;
                                ui_message = "- zoom Y";
                                break;                    
                            //------------------------------------------------
                            case 'd':
                                ail.view_scale *= 1.01;
                                ui_message = "+ zoom XY";
                                break;                    
                            //------------------------------------------------
                            case 'D':
                                ail.view_scale *= 0.99;
                                ui_message = "- zoom XY";
                                break;                    
                            //------------------------------------------------
                            case 'f':
                            case 'F':
                                ail.view_scale = 1.0;
                                ui_message = "no zoom";
                                break;
                            //------------------------------------------------
                            case 'o':
                                ail.current_frame().rotate_around_origin();
                                ail.view_angle = 0.0;
                                ui_message = "set rotate frame";
                                break;                    
                            //------------------------------------------------
                            case 'O':
                                ail.display("rotating all");
                                ail.display_please_wait();
                                ail.rotate_around_origin();
                                ail.view_angle = 0.0;
                                ui_message = "set rotate all";
                                break;                    
                            //------------------------------------------------
                            case 't':
                                ail.current_frame().move();
                                ail.view_offset = 0.0;                
                                ui_message = "set offset frame";
                                break;                    
                            //------------------------------------------------
                            case 'T':
                                ail.display("offsetting all");
                                ail.display_please_wait();
                                ail.move();
                                ail.view_offset = 0.0;                
                                ui_message = "set offset all";
                                break;                    
                            //------------------------------------------------
                            case 'g':
                                ail.current_frame().scale_around_origin();
                                ail.view_scale = 1.0;
                                ui_message = "set scale frame";
                                break;                    
                            //------------------------------------------------
                            case 'G':
                                ail.display("scaling all");
                                ail.display_please_wait();
                                ail.scale_around_origin();
                                ail.view_scale = 1.0;
                                ui_message = "set scale all";
                                break;                    
                            //------------------------------------------------
                            case 'h':
                                ail.current_frame().scale_around_origin();
                                ail.current_frame().rotate_around_origin();
                                ail.current_frame().move();        
                                ail.view_angle  = 0.0;
                                ail.view_scale  = 1.0;
                                ail.view_offset = 0.0;
                                ui_message = "set view frame";
                                break;                    
                            //------------------------------------------------
                            case 'H':
                                ail.display("setting view all");
                                ail.display_please_wait();
                                ail.scale_around_origin();
                                ail.rotate_around_origin();
                                ail.move();        
                                ail.view_angle  = 0.0;
                                ail.view_scale  = 1.0;
                                ail.view_offset = 0.0;
                                ui_message = "set view all";
                                break;                    
                            //------------------------------------------------
                            case ';': // toggle bitmap background
                                ail.show_background_bitmap = !ail.show_background_bitmap;
                                break;                    
                            //------------------------------------------------
                        }
                    }
                    ail.display(ui_message);
                    ail.display_veiw_menu();                    
                }
                break;
            }
            //----------------------------------------------------------------                                        





            //----------------------------------------------------------------
            case 'p': // Palette transformations
            case 'P':
            {
                bool show_help_was = show_help;
                show_help = true;
                in_sub_menu = true;
                ui_message  = "palette transforms";
                ail.display(ui_message);
                if(show_help)                
                    ail.display_palette_tranz_menu();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "exit palette transforms";
                                break;                                                               
                            //------------------------------------------------
                            default:
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------
                                    case SDLK_UP: // display next palette
                                        ail.target_palette_index++;
                                        if(ail.target_palette_index >= ail.number_of_palettes())
                                            ail.target_palette_index = 0;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_DOWN: // display previous palette
                                        ail.target_palette_index--;
                                        if(ail.target_palette_index < 0)
                                            ail.target_palette_index = ail.number_of_palettes() - 1;
                                        SDL_Delay(200);
                                        break;                    
                                    //----------------------------------------
                                    case SDLK_RIGHT: // display previous frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index < ail.number_of_frames() - 1)
                                          )
                                            ++ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_LEFT: // display next frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index > 0)
                                          )
                                            --ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                }
                                break;
                            //------------------------------------------------
                            case '1':
                                ail.current_frame().rainbow_recolor(0, ail.recolor_span_factor);
                                ui_message = "span X frame";
                                break;                    
                            //------------------------------------------------
                            case '!':
                                ail.display("span X all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(0, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '2':
                                ail.current_frame().rainbow_recolor(1, ail.recolor_span_factor);
                                ui_message = "span Y frame";
                                break;                    
                            //------------------------------------------------
                            case '@':
                                ail.display("span Y all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(1, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '3':
                                ail.current_frame().rainbow_recolor(2, ail.recolor_span_factor);
                                ui_message = "span Z frame";
                                break;                    
                            //------------------------------------------------
                            case '#':
                                ail.display("span Z all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(2, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '4':
                                ail.current_frame().rainbow_recolor(3, ail.recolor_span_factor);
                                ui_message = "span vertices frame";
                                break;                    
                            //------------------------------------------------
                            case '$':
                                ail.display("span vertices all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(3, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '5':
                                ail.current_frame().rainbow_recolor(4, ail.recolor_span_factor);
                                ui_message = "index vertices frame";
                                break;                    
                            //------------------------------------------------
                            case '%':
                                ail.display("index vertices all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(4, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '6':
                                ail.current_frame().rainbow_recolor(5, ail.recolor_span_factor);
                                ui_message = "span segments frame";
                                break;                    
                            //------------------------------------------------
                            case '^':
                                ail.display("span segments all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(5, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '7':
                                ail.current_frame().rainbow_recolor(6, ail.recolor_span_factor);
                                ui_message = "index segments frame";
                                break;                    
                            //------------------------------------------------
                            case '&':
                                ail.display("index segments all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(6, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '8':
                                ail.current_frame().rainbow_recolor(7, ail.recolor_span_factor);
                                ui_message = "random vertices frame";
                                break;                    
                            //------------------------------------------------
                            case '*':
                                ail.display("random vertices all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(7, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '9':
                                ail.current_frame().rainbow_recolor(8, ail.recolor_span_factor);
                                ui_message = "random segments frame";
                                break;                    
                            //------------------------------------------------
                            case '(':
                                ail.display("random segments all");
                                ail.display_please_wait();
                                ail.rainbow_recolor(8, ail.recolor_span_factor);
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case '0':
                                ail.current_frame().accept_target_palette();
                                ui_message = "by index frame";
                                break;                    
                            //------------------------------------------------
                            case ')':
                                ail.display("by index all");
                                ail.display_please_wait();
                                ail.accept_target_palette();
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case 'a':
                                ail.current_frame().best_match_target_palette();
                                ui_message = "best match frame";
                                break;
                            //------------------------------------------------
                            case 'A':
                                ail.display("best match all");
                                ail.display_please_wait();
                                ail.best_match_target_palette();
                                ui_message = "done";
                                break;
                            //------------------------------------------------
                            case 'b':
                                ail.current_frame().sync_rgb_and_palette();
                                ui_message = "24bit to palette frame";
                                break;                    
                            //------------------------------------------------
                            case 'B':
                                ail.display("24bit to palette all");
                                ail.display_please_wait();
                                ail.sync_rgb_and_palette();
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case 'c':
                                ail.current_frame().bit_reduce_to_palette();
                                ui_message = "888 to 332 frame";
                                break;                    
                            //------------------------------------------------
                            case 'C':
                                ail.display("888 to 332 all");
                                ail.display_please_wait();
                                ail.bit_reduce_to_palette();
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case 'd':
                                ail.current_frame().best_reduce_to_palette();
                                ui_message = "best reduce to palette";
                                break;                    
                            //------------------------------------------------
                            case 'D':
                                ail.display("best reduce to palette");
                                ail.display_please_wait();
                                ail.best_reduce_to_palette();
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case 'e':
                                ail.current_frame().promote_to_true_color();
                                ui_message = "promote to 24bit frame";
                                break;                    
                            //------------------------------------------------
                            case 'E':
                                ail.display("promote to 24bit all");
                                ail.display_please_wait();
                                ail.promote_to_true_color();
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case 'f':
                            case 'F':
                                ail.display("best reduce all to 1");
                                ail.display_please_wait();
                                ail. best_reduce_all_to_palette();
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case 'g':
                                ail.current_frame().strip_color();
                                ui_message = "strip color frame";
                                break;                    
                            //------------------------------------------------
                            case 'G':
                                ail.display("strip color all");
                                ail.display_please_wait();
                                ail.strip_color();
                                ui_message = "done";
                                break;                    
                            //------------------------------------------------
                            case ';': // toggle bitmap background
                                ail.show_background_bitmap = !ail.show_background_bitmap;
                                break;                    
                            //------------------------------------------------
                            case SDLK_RETURN:
                                show_help = !show_help;
                            //------------------------------------------------
                        }
                    }
                    ail.display(ui_message);
                    if(show_help)
                        ail.display_palette_tranz_menu();
                }
                show_help = show_help_was;
                break;
            }
            //----------------------------------------------------------------                                        




            //----------------------------------------------------------------
            case 'h': // frame transforms
            case 'H':
            {   // local scope                 
                bool show_help_was = show_help;
                show_help = true;
                in_sub_menu = true;
                ui_message = "frame transforms";
                ail.display(ui_message);
                if(show_help)
                    ail.display_frame_tranz_menu();
                if(show_frame_stats)
                    ail.current_frame().display_stats();
                if(show_frame_set_stats)
                    ail.display_stats();                
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "exit frame transforms";
                                break;
                            //------------------------------------------------
                            default:
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------
                                    case SDLK_RIGHT: // display previous frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index < ail.number_of_frames() - 1)
                                          )
                                            ++ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_LEFT: // display next frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index > 0)
                                          )
                                            --ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                }
                                break;
                            //------------------------------------------------
                            case '1':
                                ail.current_frame().flatten_z();
                                ui_message = "flatten Z";
                                break;                    
                            //------------------------------------------------
                            case '!':
                                ail.display("flattening Z");
                                ail.display_please_wait();                                                        
                                ail.flatten_z();
                                ui_message = "flatten Z all";
                                break;                    
                            //------------------------------------------------
                            case '2':
                                ail.current_frame().z_order_points(64000);
                                ui_message = "z order vertices";
                                break;                    
                            //------------------------------------------------
                            case '@':
                                ail.display("Z ordering vertices");
                                ail.display_please_wait();
                                ail.z_order_points(64000);
                                ui_message = "Z order vertices all";
                                break;                    
                            //------------------------------------------------
                            case '3':
                                ail.current_frame().reverse();
                                ui_message = "reverse vector order";
                                break;                    
                            //------------------------------------------------
                            case '#':
                                ail.display("reversing vector order");
                                ail.display_please_wait();
                                ail.frame_reverse();
                                ui_message = "reverse vector order all";
                                break;                    
                            //------------------------------------------------
                            case '4':
                                ail.current_frame().flip(0);
                                ui_message = "mirror";
                                break;                    
                            //------------------------------------------------
                            case '$':
                                ail.display("mirroring");
                                ail.display_please_wait();
                                ail.flip(0);
                                ui_message = "mirror all";
                                break;                    
                            //------------------------------------------------
                            case '5':
                                ail.current_frame().flip(1);
                                ui_message = "flip";
                                break;                    
                            //------------------------------------------------
                            case '%':
                                ail.display("flipping");
                                ail.display_please_wait();
                                ail.flip(1);
                                ui_message = "flip all";
                                break;                    
                            //------------------------------------------------
                            case '6':
                                ail.current_frame().quarter_turn(0, 1);
                                ui_message = "quarter turn CW";
                                break;                    
                            //------------------------------------------------
                            case '^':
                                ail.display("quarter turning CW");
                                ail.display_please_wait();
                                ail.quarter_turn(0, 1);
                                ui_message = "quarter turn CW all";
                                break;                    
                            //------------------------------------------------
                            case '7':
                                ail.current_frame().quarter_turn(0, 3);
                                ui_message = "quarter turn CCW";
                                break;                    
                            //------------------------------------------------
                            case '&':
                                ail.display("quarter turning CCW");
                                ail.display_please_wait();
                                ail.quarter_turn(0, 3);
                                ui_message = "quarter turn CCW all";
                                break;                    
                            //------------------------------------------------
                            case '8':
                                ail.current_frame().accentuate_the_positive();
                                ui_message = "accentuate the positive";
                                break;                    
                            //------------------------------------------------
                            case '*':
                                ail.display("accentuating the positive");
                                ail.display_please_wait();
                                ail.accentuate_the_positive();
                                ui_message = "accentuate the positive all";
                                break;                    
                            //------------------------------------------------
                            case '9':
                                ail.current_frame().eliminate_the_negative();
                                ui_message = "eliminate the negative";
                                break;                    
                            //------------------------------------------------
                            case '(':
                                ail.display("eliminating the negative");
                                ail.display_please_wait();
                                ail.eliminate_the_negative();
                                ui_message = "eliminate the negative all";
                                break;                    
                            //------------------------------------------------
                            case '0':
                                ail.current_frame().convert_black_to_blank();
                                ui_message = "black to blank";
                                break;                    
                            //------------------------------------------------
                            case ')':
                                ail.display("converting black to blank");
                                ail.display_please_wait();
                                ail.convert_black_to_blank();
                                ui_message = "black to blank all";
                                break;                    
                            //------------------------------------------------
                            case 'a':
                                ail.current_frame().minimize_blank_points();
                                ui_message = "remove blank span";
                                break;                    
                            //------------------------------------------------
                            case 'A':
                                ail.display("minimizing blank vectors");
                                ail.display_please_wait();
                                ail.minimize_blank_points();
                                ui_message = "remove blank span all";
                                break;                    
                            //------------------------------------------------
                            case 'b':
                                ail.current_frame().remove_dwell_points();
                                ui_message = "remove angle dwell";
                                break;                    
                            //------------------------------------------------
                            case 'B':
                                ail.display("removing dwell vertices");
                                ail.display_please_wait();
                                ail.remove_dwell_points();
                                ui_message = "remove angle dwell all";
                                break;                    
                            //------------------------------------------------
                            case 'c':
                                ail.current_frame().minimize_vectors();
                                ui_message = "reduce vectors";
                                break;                    
                            //------------------------------------------------
                            case 'C':
                                ail.display("reducing vectors");
                                ail.display_please_wait();
                                ail.minimize_vectors();
                                ui_message = "reduce vectors all";
                                break;                    
                            //------------------------------------------------
                            case 'd':
                                ail.current_frame().blank_equivelant_segments();
                                ui_message = "blank equal segments";
                                break;                    
                            //------------------------------------------------
                            case 'D':
                                ail.display("blanking equal segments");
                                ail.display_please_wait();
                                ail.blank_equivelant_segments();
                                ui_message = "blank equal segments";
                                break;                    
                            //------------------------------------------------
                            case 'e':
                                ail.current_frame().bond_segments();
                                ui_message = "bond segments";
                                break;                    
                            //------------------------------------------------
                            case 'E':
                                ail.display("bonding segments");
                                ail.display_please_wait();
                                ail.bond_segments();
                                ui_message = "bond segments all";
                                break;                    
                            //------------------------------------------------
                            case 'f':
                                ail.current_frame().minimize_points();
                                ui_message = "minimize vectors";
                                break;                    
                            //------------------------------------------------
                            case 'F':
                                ail.display("minimizing vectors");
                                ail.display_please_wait();
                                ail.minimize_points();
                                ui_message = "minimize vectors all";
                                break;                    
                            //------------------------------------------------
                            case 'g':
                                ail.current_frame().fracture_segments();
                                ui_message = "fracture segments";
                                break;                    
                            //------------------------------------------------
                            case 'G':
                                ail.display("fracturing segments");
                                ail.display_please_wait();
                                ail.fracture_segments();
                                ui_message = "fracture segments all";
                                break;                    
                            //------------------------------------------------
                            case 'h':
                                ail.current_frame().reorder_segments();
                                ui_message = "reorder segments";
                                break;                    
                            //------------------------------------------------
                            case 'H':
                                ail.display("reordering segments");
                                ail.display_please_wait();
                                ail.reorder_segments();
                                ui_message = "reorder segments all";
                                break;                    
                            //------------------------------------------------
                            case 'i':
                                ail.current_frame().add_dwell_points();
                                ui_message = "add angle dwell";
                                break;                    
                            //------------------------------------------------
                            case 'I':
                                ail.display("adding angle dwell");
                                ail.display_please_wait();
                                ail.add_dwell_points();
                                ui_message = "add angle dwell all";
                                break;                    
                            //------------------------------------------------
                            case 'j':
                                ail.current_frame().add_lit_span_points();
                                ui_message = "add lit distance span";
                                break;                    
                            //------------------------------------------------
                            case 'J':
                                ail.display("adding lit distance span");
                                ail.display_please_wait();
                                ail.add_lit_span_points();
                                ui_message = "add lit distance span all";
                                break;                    
                            //------------------------------------------------
                            case 'k':
                                ail.current_frame().add_blank_span_points();
                                ui_message = "add blank distance span";
                                break;                    
                            //------------------------------------------------
                            case 'K':
                                ail.display("adding blank distance span");
                                ail.display_please_wait();
                                ail.add_blank_span_points();
                                ui_message = "add blank distance span all";
                                break;                    
                            //------------------------------------------------
                            case 'l':
                                ail.current_frame().optimize_points();
                                ui_message = "optimize";
                                break;                    
                            //------------------------------------------------
                            case 'L':
                                ail.display("optimizing");
                                ail.display_please_wait();
                                ail.optimize_points();
                                ui_message = "optimize all";
                                break;                    
                            //------------------------------------------------
                            case ';': // toggle bitmap background
                                ail.show_background_bitmap = !ail.show_background_bitmap;
                                break;                    
                            //------------------------------------------------
                            case SDLK_RETURN:
                                show_help = !show_help;
                            //------------------------------------------------
                        }
                    }
                    ail.display(ui_message);
                    if(show_frame_stats)
                        ail.current_frame().display_stats();
                    if(show_frame_set_stats)
                        ail.display_stats();
                    if(show_help)
                        ail.display_frame_tranz_menu();
                }
                show_help = show_help_was;
                break;
            }   // end local scope
            //----------------------------------------------------------------





            //----------------------------------------------------------------
            case 'j': // frame set transforms
            case 'J':
            {   // local scope
                bool show_help_was = show_help;
                show_help = true;
                in_sub_menu = true;
                ui_message = "frame set transforms";
                ail.display(ui_message);
                if(show_help)
                    ail.display_frame_set_tranz_menu();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "action canceled";
                                break;
                            //------------------------------------------------
                            default:
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------
                                    case SDLK_RIGHT: // display previous frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index < ail.number_of_frames() - 1)
                                          )
                                            ++ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_LEFT: // display next frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index > 0)
                                          )
                                            --ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                }
                                break;
                            //------------------------------------------------
                            case '1':
                                ail.display("deleting selected frames");
                                ail.display_please_wait();
                                ail.delete_selected_frames();
                                ui_message = "deleted selected frames";
                                break;
                            //------------------------------------------------
                            case '2':   
                                ail.display("trimming to selected frames");
                                ail.display_please_wait();
                                ail.trim_to_selected_frames();
                                ui_message = "trimmed to selected frames";
                                break;
                            //------------------------------------------------
                            case '3':
                                ail.display("deleting redundant frames");
                                ail.display_please_wait();
                                ail.delete_redundant_frames();
                                ui_message = "redundant frames deleted";
                                break;
                            //------------------------------------------------
                            case '4':
                                ail.display("copying selected frames");
                                ail.display_please_wait();
                                ail = (ail.selected_frames() + ail);
                                ui_message = "copied frames to begining";
                                break;
                            //------------------------------------------------
                            case '5':
                                ail.display("copying selected frames");
                                ail.display_please_wait();
                                ail += ail.selected_frames();
                                ui_message = "copied frames to end";
                                break;
                            //------------------------------------------------
                            case '6':
                                ail.display("moving selected frames");
                                ail.display_please_wait();
                                {   
                                    ez_uail_frame_set temp(ail.selected_frames());
                                    ail.delete_selected_frames();
                                    temp += ail;
                                    ail = temp;
                                }
                                ui_message = "moved frames to begining";
                                break;
                            //------------------------------------------------
                            case '7':
                                ail.display("moving selected frames");
                                ail.display_please_wait();
                                {
                                    ez_uail_frame_set temp(ail.selected_frames());
                                    ail.delete_selected_frames();
                                    ail += temp;
                                }
                                ui_message = "moved frames to end";
                                break;
                            //------------------------------------------------
                            case '8':
                                {
                                    ez_uail_frame_set temp;
                                    temp += ez_uail_frame(ail.target_palette_index, 0); // not quite an empty frame!
                                    ail = temp + ail;
                                }
                                ui_message = "add frame to begining";
                                break;
                            //------------------------------------------------
                            case '9':
                                {
                                    int               i;
                                    ez_uail_frame_set temp;
                                    for(i = 0; i <= ail.frame_index; i++)
                                        temp += ail[i];
                                    temp += ez_uail_frame(ail.target_palette_index, 0);
                                    for(i = ail.frame_index + 1; i < ail.number_of_frames(); i++)
                                        temp += ail[i];
                                    ail = temp;
                                }
                                ui_message = "add after current frame";
                                break;
                            //------------------------------------------------
                            case '0':
                                ail += ez_uail_frame(ail.target_palette_index, 0);
                                ui_message = "add frame to end";
                                break;
                            //------------------------------------------------
                            case 'a':
                                ail.display("reversing order of frames");
                                ail.display_please_wait();
                                ail.reverse();
                                ui_message = "reverse order";
                                break;
                            //------------------------------------------------
                            case 'b':
                                ail.display("random order frames");
                                ail.display_please_wait();
                                ail.random_order();
                                ui_message = "random order";
                                break;
                            //------------------------------------------------
                            case '>': // explode current frame
                                ail.explode_current_frame();
                                ui_message = "explode segments";
                                break;
                            //------------------------------------------------
                            case '<': // collapse selected frames
                                ail.collapse_selected_frames();
                                ui_message = "collapse segments";
                                break;
                            //------------------------------------------------
                            case '!': // explode all frames
                                ail.explode_all_frames();
                                ui_message = "explode segments";
                                break;
                            //------------------------------------------------
                            case ' ': // [Space] select, deselect current frame
                                ail.current_frame().is_selected = !(ail.current_frame().is_selected);
                                if(ail.current_frame().is_selected)
                                    ui_message = "select frame";
                                else
                                    ui_message = "deselect frame";
                                break;
                            //------------------------------------------------
                            case ';': // toggle bitmap background
                                ail.show_background_bitmap = !ail.show_background_bitmap;
                                break;                    
                            //------------------------------------------------
                            case SDLK_RETURN:
                                show_help = !show_help;
                            //------------------------------------------------
                        }
                    }
                    ail.display(ui_message);
                    if(show_help)
                        ail.display_frame_set_tranz_menu();
                }
                break;
            } // end local scope
            //----------------------------------------------------------------





            //----------------------------------------------------------------
            case 'k': // draw, color, blank
            case 'K':
            {   // local scope
                int         number_of_sides;
                float       temp_float;
                ez_3D_float temp_3D_float;
                
                bool show_points_was   = ail.show_points,              
                     show_blanking_was = ail.show_blanking,
                     show_help_was     = show_help;

                show_help         = true;
                ail.show_points   = true;
                ail.show_blanking = true;
                ail.show_cursors  = true;
                in_sub_menu       = true;
                ui_message = "draw color blank";                
                ail.display(ui_message);
                if(show_help)
                    ail.display_draw_color_blank_menu();                        
                if(show_frame_stats)
                    ail.current_frame().display_stats();
                if(show_frame_set_stats)
                    ail.display_stats();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu       = false;
                                ail.show_cursors  = false;
                                ail.show_points   = show_points_was;
                                ail.show_blanking = show_blanking_was;                                
                                ui_message = "exit draw color blank";
                                break;
                            //------------------------------------------------
                            case SDLK_BACKSPACE:
                                ail.current_frame().remove_point(ail.current_frame().point_select_end);
                                ui_message = "delete vertex";
                                break;
                            //------------------------------------------------
                            default:
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------
                                    case SDLK_UP: // display next palette
                                        ail.target_palette_index++;
                                        if(ail.target_palette_index >= ail.number_of_palettes())
                                            ail.target_palette_index = 0;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_DOWN: // display previous palette
                                        ail.target_palette_index--;
                                        if(ail.target_palette_index < 0)
                                            ail.target_palette_index = ail.number_of_palettes() - 1;
                                        SDL_Delay(200);
                                        break;                    
                                    //----------------------------------------
                                    case SDLK_RIGHT: // display previous frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index < ail.number_of_frames() - 1)
                                          )
                                            ++ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_LEFT: // display next frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index > 0)
                                          )
                                            --ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                }
                                break;
                            //------------------------------------------------
                            case '1':
                                ail.view_angle.x += ail.rotate_radians_per_key;
                                ui_message = "+ orbit X";
                                break;                    
                            //------------------------------------------------
                            case '!':
                                ail.view_angle.x -= ail.rotate_radians_per_key;
                                ui_message = "- orbit X";
                                break;                    
                            //------------------------------------------------
                            case '2':
                                ail.view_angle.y += ail.rotate_radians_per_key;
                                ui_message = "+ orbit Y";
                                break;                    
                            //------------------------------------------------
                            case '@':
                                ail.view_angle.y -= ail.rotate_radians_per_key;
                                ui_message = "- orbit Y";
                                break;                    
                            //------------------------------------------------
                            case '3':
                                ail.view_angle.z += ail.rotate_radians_per_key;
                                ui_message = "+ orbit Z";
                                break;                    
                            //------------------------------------------------
                            case '#':
                                ail.view_angle.z -= ail.rotate_radians_per_key;
                                ui_message = "- orbit Z";
                                break;                    
                            //------------------------------------------------
                            case '4':
                                ail.view_angle += ail.rotate_radians_per_key;
                                ui_message = "+ orbit XYZ";
                                break;                    
                            //------------------------------------------------
                            case '$':
                                ail.view_angle -= ail.rotate_radians_per_key;
                                ui_message = "- orbit XYZ";
                                break;                    
                            //------------------------------------------------
                            case '5':
                                ail.view_angle   = 0;
                                ui_message = "front";
                                break;                    
                            //------------------------------------------------
                            case '%':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = pi;
                                ail.view_angle.z = 0;
                                ui_message = "back";
                                break;                    
                            //------------------------------------------------
                            case '6':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = half_pi;
                                ail.view_angle.z = 0;
                                ui_message = "left side";
                                break;                    
                            //------------------------------------------------
                            case '^':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = -half_pi;
                                ail.view_angle.z = 0;
                                ui_message = "right side";
                                break;                    
                            //------------------------------------------------
                            case '7':
                                ail.view_angle.x = half_pi;
                                ail.view_angle.y = 0;
                                ail.view_angle.z = 0;
                                ui_message = "top";
                                break;                    
                            //------------------------------------------------
                            case '&':
                                ail.view_angle.x = -half_pi;
                                ail.view_angle.y = 0;
                                ail.view_angle.z = 0;
                                ui_message = "bottom";
                                break;                    
                            //------------------------------------------------
                            case '8':
                                ail.view_angle += ail.rotate_radians_per_key;
                                if(    ail.loop_animation
                                    || (ail.frame_index < ail.number_of_frames() - 1)
                                  )
                                    ++ail;
                                ui_message = "+ frame & orbit";
                                break;
                            //------------------------------------------------
                            case '*':
                                ail.view_angle -= ail.rotate_radians_per_key;
                                if(    ail.loop_animation
                                    || (ail.frame_index > 0)
                                  )
                                    --ail;
                                ui_message = "- frame & orbit";
                                break;
                            //------------------------------------------------
                            case '[':
                                ail.current_frame().increment_point_select_start();
                                ui_message = "+ start vertex select";
                                break;                    
                            //------------------------------------------------
                            case '{':
                                ail.current_frame().deccrement_point_select_start();
                                ui_message = "- start vertex select";
                                break;                    
                            //------------------------------------------------
                            case ']':
                                ail.current_frame().increment_point_select_end();
                                ui_message = "+ end vertex select";
                                break;                    
                            //------------------------------------------------
                            case '}':
                                ail.current_frame().deccrement_point_select_end();
                                ui_message = "- end vertex select";
                                break;                    
                            //------------------------------------------------
                            case'\\':
                                ail.current_frame().point_select_end = (ail.current_frame().number_of_points() - 1);
                                ui_message = "last vertex select";
                                break;                    
                            //------------------------------------------------
                            case '|':
                                ail.current_frame().point_select_start = 0;
                                ui_message = "first vertex select";
                                break;                    
                            //------------------------------------------------
                            case 'u':
                                ail.current_frame().next_segment_select();
                                ui_message = "+ lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'U':
                                ail.current_frame().previous_segment_select();
                                ui_message = "- lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'i':
                                ail.current_frame().next_segment_select_start();
                                ui_message = "+ start lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'I':
                                ail.current_frame().previous_segment_select_start();
                                ui_message = "- start lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'o':
                                ail.current_frame().next_segment_select_end();
                                ui_message = "+ end lit segment";
                                break;
                            //------------------------------------------------
                            case 'O':
                                ail.current_frame().previous_segment_select_end();
                                ui_message = "- end lit segment";
                                break;
                            //------------------------------------------------
                            case '/': // show frame set stats
                                show_frame_set_stats = !show_frame_set_stats;
                                ui_message = "frame set stats";
                                break;                    
                            //------------------------------------------------
                            case '?': // show frame stats
                                show_frame_stats = !show_frame_stats;
                                ui_message = "frame stats";
                                break;
                            //------------------------------------------------
                            case 'q':
                                temp_int = ail.display_prompt_and_echo_int("set value of X");
                                if(ail.process_exception)
                                    break;
                                ail.current_frame().at(ail.current_frame().point_select_end).x = temp_int;
                                ui_message = "X set";
                                break;
                            //------------------------------------------------
                            case 'w':
                                temp_int = ail.display_prompt_and_echo_int("set value of Y");
                                if(ail.process_exception)
                                    break;
                                ail.current_frame().at(ail.current_frame().point_select_end).y = temp_int;
                                ui_message = "Y set";
                                break;
                            //------------------------------------------------
                            case 'e':
                                temp_int = ail.display_prompt_and_echo_int("set value of Z");
                                if(ail.process_exception)
                                    break;
                                ail.current_frame().at(ail.current_frame().point_select_end).z = temp_int;
                                ui_message = "Z set";
                                break;
                            //------------------------------------------------
                            case 'p':
                                ail.selected_palette_color++;
                                if(ail.selected_palette_color >= ail.current_palette().number_of_colors())
                                    ail.selected_palette_color = 0;
                                ui_message = "+ color index";
                                break;                    
                            //------------------------------------------------
                            case 'P':
                                ail.selected_palette_color--;
                                if(ail.selected_palette_color < 0)
                                    ail.selected_palette_color = ail.current_palette().number_of_colors() - 1;
                                ui_message = "- color index";
                                break;                    
                            //------------------------------------------------
                            case 'a':
                                ail.current_frame().unblank_all_points();
                                ui_message = "unblank all vectors";
                                break;                    
                            //------------------------------------------------
                            case 'A':
                                ail.unblank_all_points();
                                ui_message = "unblank all frames";
                                break;                    
                            //------------------------------------------------
                            case 'b':
                                ail.current_frame().blank_points();
                                ui_message = "blank selected vectors";
                                break;                    
                            //------------------------------------------------
                            case 'B':
                                ail.current_frame().unblank_points();
                                ui_message = "unblank selected vectors";
                                break;                    
                            //------------------------------------------------
                            case 'n':
                                ail.current_frame().black_points();
                                ui_message = "bond black selected vectors";
                                break;                    
                            //------------------------------------------------
                            case 'N':
                                ail.current_frame().unblack_points();
                                ui_message = "unbond unblack selected vectors";
                                break;                    
                            //------------------------------------------------
                            case 'c':
                                ail.current_frame().set_points_color_index();
                                ui_message = "color selected vectors";
                                break;                    
                            //------------------------------------------------
                            case 'C':
                                ail.selected_palette_color = ail.current_frame().at(ail.current_frame().point_select_end).c;
                                ui_message = "get color";
                                break;                    
                            //------------------------------------------------
                            case '.':
                                ail.current_frame().add_new_point(ail.current_frame().point_select_end);
                                if(ail.current_frame().point_select_end < (ail.current_frame().number_of_points() - 1))
                                    ui_message = "split vector";
                                else
                                    ui_message = "add vector";
                                break;
                            //------------------------------------------------
                            case 't':
                                {
                                    float  x_rotation;
                                    string font,
                                           text;
                                    //list_uail_files();
                                    font = ail.display_prompt_and_echo_name("enter font (ILDA) file name");
                                    if(ail.process_exception)
                                        break;
                                    ail.display(ui_message);
                                    if(show_help)
                                        ail.display_draw_color_blank_menu();
                                    x_rotation = ail.display_prompt_and_echo_float ("enter angle of rotation around vector");
                                    if(ail.process_exception)
                                        break;
                                    ail.display(ui_message);
                                    if(show_help)
                                        ail.display_draw_color_blank_menu();
                                    text = ail.display_prompt_and_echo_string("text to render");
                                    if(ail.process_exception)
                                        break;
                                    ail.display("rendering text");
                                    ail.display_please_wait();
                                    ail.current_frame()
                                    += ez_uail_segment(   ail.current_frame().at(ail.current_frame().point_select_start)
                                                        , ail.current_frame().at(ail.current_frame().point_select_end)
                                                        , x_rotation
                                                        , font
                                                        , text
                                                      );
                                    ail.current_frame().at(ail.current_frame().point_select_start).blank();
                                    ui_message = "text rendered";
                                }
                                break;
                            //------------------------------------------------
                            case 'k':
                                {
                                    int number_of_sides = ail.display_prompt_and_echo_int("enter number of sides");
                                    if(ail.process_exception)
                                        break;
                                    ail.current_frame()
                                        += ez_uail_segment(   ail.current_frame().at(ail.current_frame().point_select_start)
                                                            , ail.current_frame().at(ail.current_frame().point_select_end)
                                                            , number_of_sides
                                                          );
                                }
                                break;
                            //------------------------------------------------
                            case 'd':
                                {
                                    float   arc_angle,
                                            radii_ratio;

                                    arc_angle = ail.display_prompt_and_echo_float("enter circular arc angle in degrees, (0 = 360)"); 
                                    if(ail.process_exception)
                                        break;
                                    ail.display(ui_message);
                                    if(show_help)
                                        ail.display_draw_color_blank_menu();
                                    if(ail.current_frame().number_of_points() > 1)
                                    {
                                        ail.current_frame()
                                        += ez_uail_segment(   ail.current_frame().at(ail.current_frame().point_select_start)
                                                            , ail.current_frame().at(ail.current_frame().point_select_end)
                                                            , arc_angle
                                                          );
                                        ui_message = "circular arc";
                                    }
                                    else
                                        ui_message = "All Your Base Are Belong To Us!";
                                }
                                break;
                            //------------------------------------------------
                            case 'D':
                                {
                                    float   arc_angle,
                                            radii_ratio;

                                    arc_angle = ail.display_prompt_and_echo_float("enter elliptical arc angle in degrees, (0 = 360)"); 
                                    if(ail.process_exception)
                                        break;
                                    ail.display(ui_message);
                                    if(show_help)
                                        ail.display_draw_color_blank_menu();                        
                                    radii_ratio = ail.display_prompt_and_echo_float("enter minor / major radii ratio");
                                    if(ail.process_exception)
                                        break;
                                    ail.display(ui_message);
                                    if(show_help)
                                        ail.display_draw_color_blank_menu();                        
                                    ail.current_frame()
                                        += ez_uail_segment(   ail.current_frame().at(ail.current_frame().point_select_start)
                                                            , ail.current_frame().at(ail.current_frame().point_select_end)
                                                            , arc_angle
                                                            , radii_ratio
                                                          );
                                    ui_message = "elliptical arc";
                                }
                                break;
                            //------------------------------------------------
                            case 'l':
                                ail.current_frame().connect_the_dots();
                                ui_message = "draw line";
                                break;                                
                            //------------------------------------------------
                            case 's':
                                ail.current_frame().break_segment();
                                ui_message = "break segment";
                                break;
                            //------------------------------------------------
                            case 'x':
                                temp_float = ail.current_frame().at(ail.current_frame().point_select_end).x + ail.move_points_per_key;
                                if(temp_float <= EZ_MAX_SHORT)
                                    ail.current_frame().at(ail.current_frame().point_select_end).x = (u_short)temp_float;
                                ui_message = "+ X";
                                break;                    
                            //------------------------------------------------
                            case 'X':
                                temp_float = ail.current_frame().at(ail.current_frame().point_select_end).x - ail.move_points_per_key;
                                if(temp_float >= EZ_MIN_SHORT)
                                    ail.current_frame().at(ail.current_frame().point_select_end).x = (u_short)temp_float;
                                ui_message = "- X";
                                break;                    
                            //------------------------------------------------
                            case 'y':
                                temp_float = ail.current_frame().at(ail.current_frame().point_select_end).y + ail.move_points_per_key;
                                if(temp_float <= EZ_MAX_SHORT)
                                    ail.current_frame().at(ail.current_frame().point_select_end).y = (u_short)temp_float;
                                ui_message = "+ Y";
                                break;                    
                            //------------------------------------------------
                            case 'Y':
                                temp_float = ail.current_frame().at(ail.current_frame().point_select_end).y - ail.move_points_per_key;
                                if(temp_float >= EZ_MIN_SHORT)
                                    ail.current_frame().at(ail.current_frame().point_select_end).y = (u_short)temp_float;
                                ui_message = "- Y";
                                break;                    
                            //------------------------------------------------
                            case 'z':
                                temp_float = ail.current_frame().at(ail.current_frame().point_select_end).z + ail.move_points_per_key;
                                if(temp_float <= EZ_MAX_SHORT)
                                    ail.current_frame().at(ail.current_frame().point_select_end).z = (u_short)temp_float;
                                ui_message = "+ Z";
                                break;                    
                            //------------------------------------------------
                            case 'Z':
                                temp_float = ail.current_frame().at(ail.current_frame().point_select_end).z - ail.move_points_per_key;
                                if(temp_float >= EZ_MIN_SHORT)
                                    ail.current_frame().at(ail.current_frame().point_select_end).z = (u_short)temp_float;
                                ui_message = "- Z";
                                break;                    
                            //------------------------------------------------
                            case 'f':
                                if(ail.current_frame().point_select_end) // is not the origin
                                {
                                   temp_3D_float
                                    = rotate_point_around_point_z(   ail.current_frame().at(ail.current_frame().point_select_end)
                                                                   , ail.current_frame().at(ail.current_frame().point_select_end - 1)
                                                                   , ail.rotate_radians_per_key
                                                                 );
                                    if(!ez_uail_bounds_check(temp_3D_float, EZ_UAIL_CUBE))
                                        ail.current_frame().at(ail.current_frame().point_select_end) = temp_3D_float;
                                    ui_message = "+ rotate on Z axis";
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //------------------------------------------------
                            case 'F':
                                if(ail.current_frame().point_select_end) // is not the origin
                                {

                                    temp_3D_float
                                    = rotate_point_around_point_z(   ail.current_frame().at(ail.current_frame().point_select_end)
                                                                   , ail.current_frame().at(ail.current_frame().point_select_end - 1)
                                                                   , -ail.rotate_radians_per_key
                                                                 );
                                    if(!ez_uail_bounds_check(temp_3D_float, EZ_UAIL_CUBE))
                                        ail.current_frame().at(ail.current_frame().point_select_end) = temp_3D_float;
                                    ui_message = "- rotate on Z axis";
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //------------------------------------------------
                            case 'g':
                                if(ail.current_frame().point_select_end) // is not the origin
                                {
                                    temp_3D_float
                                    = rotate_point_around_point_y(   ail.current_frame().at(ail.current_frame().point_select_end)
                                                                   , ail.current_frame().at(ail.current_frame().point_select_end - 1)
                                                                   , ail.rotate_radians_per_key
                                                                 );
                                    if(!ez_uail_bounds_check(temp_3D_float, EZ_UAIL_CUBE))
                                        ail.current_frame().at(ail.current_frame().point_select_end) = temp_3D_float;
                                    ui_message = "+ rotate on Y axis";
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //------------------------------------------------
                            case 'G':
                                if(ail.current_frame().point_select_end) // is not the origin
                                {
                                    temp_3D_float
                                    = rotate_point_around_point_y(   ail.current_frame().at(ail.current_frame().point_select_end)
                                                                   , ail.current_frame().at(ail.current_frame().point_select_end - 1)
                                                                   , -ail.rotate_radians_per_key
                                                                 );
                                    if(!ez_uail_bounds_check(temp_3D_float, EZ_UAIL_CUBE))
                                        ail.current_frame().at(ail.current_frame().point_select_end) = temp_3D_float;
                                    ui_message = "- rotate on Y axis";
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //------------------------------------------------
                            case 'm':
                                if(ail.current_frame().point_select_end) // is not the origin
                                {
                                temp_3D_float
                                = scale_point_around_point(   ail.current_frame().at(ail.current_frame().point_select_end)
                                                            , ail.current_frame().at(ail.current_frame().point_select_end - 1)
                                                            , ez_3D_float(1.01, 1.01, 1.01)
                                                          );
                                if(!ez_uail_bounds_check(temp_3D_float, EZ_UAIL_CUBE))
                                    ail.current_frame().at(ail.current_frame().point_select_end) = temp_3D_float;
                                    ui_message = "+ magnitude of vector";
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //------------------------------------------------
                            case 'M':
                                if(ail.current_frame().point_select_end) // is not the origin
                                {
                                    temp_3D_float                                          
                                    = scale_point_around_point(   ail.current_frame().at(ail.current_frame().point_select_end)
                                                                , ail.current_frame().at(ail.current_frame().point_select_end - 1)
                                                                , ez_3D_float(0.99, 0.99, 0.99)
                                                              );
                                    if(!ez_uail_bounds_check(temp_3D_float, EZ_UAIL_CUBE))
                                        ail.current_frame().at(ail.current_frame().point_select_end) = temp_3D_float;
                                    ui_message = "- magnitude of vector";
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //------------------------------------------------
                            case ';': // toggle bitmap background
                                ail.show_background_bitmap = !ail.show_background_bitmap;
                                break;                    
                            //------------------------------------------------
                            case SDLK_RETURN:
                                show_help = !show_help;
                            //------------------------------------------------
                        } // end switch(event.key.keysym.unicode)
                        ail.display(ui_message);
                        if(show_help)
                            ail.display_draw_color_blank_menu();                        
                        if(show_frame_stats)
                            ail.current_frame().display_stats();
                        if(show_frame_set_stats)
                            ail.display_stats();
                    } // end else if(event.type == SDL_KEYDOWN)
                } // end while(in_sub_menu)
                show_help = show_help_was;
                break;
            }   // end local scope
            //----------------------------------------------------------------






            //----------------------------------------------------------------
            case 'l': // move scale rotate
            case 'L':
            {   // local scope
                u_short offset;
                bool show_points_was   = ail.show_points,              
                     show_blanking_was = ail.show_blanking,
                     show_help_was     = show_help;

                show_help                = true;
                ail.show_points          = true;
                ail.show_blanking        = true;
                ail.show_cursors         = true;
                ail.show_arbitrary_point = true;
                in_sub_menu              = true;
                ui_message               = "move scale rotate";
                ail.display(ui_message);
                if(show_help)
                    ail.display_move_scale_rotate_menu();             
                if(show_frame_stats)
                    ail.current_frame().display_stats();
                if(show_frame_set_stats)
                    ail.display_stats();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu              = false;
                                ail.show_points          = show_points_was;
                                ail.show_blanking        = show_blanking_was;
                                ail.show_cursors         = false;
                                ail.show_arbitrary_point = false;
                                ui_message = "exit move scale rotate";
                                break;
                            //------------------------------------------------
                            default:
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------
                                    case SDLK_UP: // display next palette
                                        ail.target_palette_index++;
                                        if(ail.target_palette_index >= ail.number_of_palettes())
                                            ail.target_palette_index = 0;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_DOWN: // display previous palette
                                        ail.target_palette_index--;
                                        if(ail.target_palette_index < 0)
                                            ail.target_palette_index = ail.number_of_palettes() - 1;
                                        SDL_Delay(200);
                                        break;                    
                                    //----------------------------------------
                                    case SDLK_RIGHT: // display previous frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index < ail.number_of_frames() - 1)
                                          )
                                            ++ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                    case SDLK_LEFT: // display next frame
                                        if(    ail.loop_animation
                                            || (ail.frame_index > 0)
                                          )
                                            --ail;
                                        SDL_Delay(200);
                                        break;
                                    //----------------------------------------
                                }
                                break;
                            //------------------------------------------------
                            case '1':
                                ail.view_angle.x += ail.rotate_radians_per_key;
                                ui_message = "+ orbit X";
                                break;                    
                            //------------------------------------------------
                            case '!':
                                ail.view_angle.x -= ail.rotate_radians_per_key;
                                ui_message = "- orbit X";
                                break;                    
                            //------------------------------------------------
                            case '2':
                                ail.view_angle.y += ail.rotate_radians_per_key;
                                ui_message = "+ orbit Y";
                                break;                    
                            case '@':
                                ail.view_angle.y -= ail.rotate_radians_per_key;
                                ui_message = "- orbit Y";
                                break;                    
                            //------------------------------------------------
                            case '3':
                                ail.view_angle.z += ail.rotate_radians_per_key;
                                ui_message = "+ orbit Z";
                                break;                    
                            //------------------------------------------------
                            case '#':
                                ail.view_angle.z -= ail.rotate_radians_per_key;
                                ui_message = "- orbit Z";
                                break;                    
                            //------------------------------------------------
                            case '4':
                                ail.view_angle += ail.rotate_radians_per_key;
                                ui_message = "+ orbit XYZ";
                                break;                    
                            //------------------------------------------------
                            case '$':
                                ail.view_angle -= ail.rotate_radians_per_key;
                                ui_message = "- orbit XYZ";
                                break;                    
                            //------------------------------------------------
                            case '5':
                                ail.view_angle   = 0;
                                ui_message = "front";
                                break;                    
                            //------------------------------------------------
                            case '%':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = pi;
                                ail.view_angle.z = 0;
                                ui_message = "back";
                                break;                    
                            //------------------------------------------------
                            case '6':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = half_pi;
                                ail.view_angle.z = 0;
                                ui_message = "left side";
                                break;                    
                            //------------------------------------------------
                            case '^':
                                ail.view_angle.x = 0;
                                ail.view_angle.y = -half_pi;
                                ail.view_angle.z = 0;
                                ui_message = "right side";
                                break;                    
                            //------------------------------------------------
                            case '7':
                                ail.view_angle.x = half_pi;
                                ail.view_angle.y = 0;
                                ail.view_angle.z = 0;
                                ui_message = "top";
                                break;                    
                            //------------------------------------------------
                            case '&':
                                ail.view_angle.x = -half_pi;
                                ail.view_angle.y = 0;
                                ail.view_angle.z = 0;
                                ui_message = "bottom";
                                break;                    
                            //------------------------------------------------
                            case '8':
                                ail.view_angle += ail.rotate_radians_per_key;
                                if(    ail.loop_animation
                                    || (ail.frame_index < ail.number_of_frames() - 1)
                                  )
                                    ++ail;
                                ui_message = "+ show & orbit";
                                break;
                            //------------------------------------------------
                            case '*':
                                ail.view_angle -= ail.rotate_radians_per_key;
                                if(    ail.loop_animation
                                    || (ail.frame_index > 0)
                                  )
                                    --ail;
                                ui_message = "- show & orbit";
                                break;
                            //------------------------------------------------
                            case '[':
                                ail.current_frame().increment_point_select_start();
                                ui_message = "+ start vertex  select";
                                break;                    
                            //------------------------------------------------
                            case '{':
                                ail.current_frame().deccrement_point_select_start();
                                ui_message = "- start vertex select";
                                break;                    
                            //------------------------------------------------
                            case ']':
                                ail.current_frame().increment_point_select_end();
                                ui_message = "+ end vertex  select";
                                break;                    
                            //------------------------------------------------
                            case '}':
                                ail.current_frame().deccrement_point_select_end();
                                ui_message = "- end vertex  select";
                                break;                    
                            //------------------------------------------------
                            case'\\':
                                ail.current_frame().point_select_end = (ail.current_frame().number_of_points() - 1);
                                ui_message = "last vertex select";
                                break;                    
                            //------------------------------------------------
                            case '|':
                                ail.current_frame().point_select_start = 0;
                                ui_message = "first vertex select";
                                break;                    
                            //------------------------------------------------
                            case 'u':
                                ail.current_frame().next_segment_select();
                                ui_message = "+ lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'U':
                                ail.current_frame().previous_segment_select();
                                ui_message = "- lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'i':
                                ail.current_frame().next_segment_select_start();
                                ui_message = "+ start lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'I':
                                ail.current_frame().previous_segment_select_start();
                                ui_message = "- start lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'o':
                                ail.current_frame().next_segment_select_end();
                                ui_message = "+ end lit segment";
                                break;                    
                            //------------------------------------------------
                            case 'O':
                                ail.current_frame().previous_segment_select_end();
                                ui_message = "- end lit segment";
                                break;                                                    
                            //------------------------------------------------
                            case '/': // show frame set stats
                                show_frame_set_stats = !show_frame_set_stats;
                                break;                    
                            //------------------------------------------------
                            case '?': // show frame stats
                                show_frame_stats = !show_frame_stats;
                                break;                    
                            //------------------------------------------------
                            case 'c':
                                ail.current_frame() += ail.current_frame().selected_segment();
                                sprintf(formatted_string, "copied %d vertices to end of frame", ail.current_frame().selected_segment().size());
                                ui_message = formatted_string;
                                clear_string(formatted_string);
                                break;
                            //------------------------------------------------
                            case 'j':
                                temp_float = ail.display_prompt_and_echo_float("set a point X");
                                if(ail.process_exception)
                                    break;
                                ail.arbitrary_point.x = temp_float;
                                ui_message = "point X set";
                                break;
                            //------------------------------------------------
                            case 'k':
                                temp_float = ail.display_prompt_and_echo_float("set a point Y");
                                if(ail.process_exception)
                                    break;
                                ail.arbitrary_point.y = temp_float;
                                ui_message = "point Y set";
                                break;
                            //------------------------------------------------
                            case 'l':
                                temp_float = ail.display_prompt_and_echo_float("set a point Z");
                                if(ail.process_exception)
                                    break;
                                ail.arbitrary_point.z = temp_float;
                                ui_message = "point Z set";
                                break;
                            //------------------------------------------------
                            case 'q':
                                offset = ail.display_prompt_and_echo_int("set center X");
                                if(ail.process_exception)
                                    break;
                                ail.current_frame().move_selection(ez_3D_float(   offset - ail.current_frame().selected_segment().find_center().x
                                                                                , 0
                                                                                , 0
                                                                              )
                                                                  );
                                break;
                            //------------------------------------------------
                            case 'w':
                                offset = ail.display_prompt_and_echo_int("set center Y");
                                if(ail.process_exception)
                                    break;
                                ail.current_frame().move_selection(ez_3D_float(   0
                                                                                , offset - ail.current_frame().selected_segment().find_center().y
                                                                                , 0
                                                                              )
                                                                  );
                                break;
                            //------------------------------------------------
                            case 'e':
                                offset = ail.display_prompt_and_echo_int("set center Z");
                                if(ail.process_exception)
                                    break;
                                ail.current_frame().move_selection(ez_3D_float(   0
                                                                                , 0
                                                                                , offset - ail.current_frame().selected_segment().find_center().z
                                                                              )
                                                                  );
                                break;
                            //------------------------------------------------
                            case 'x':
                                ail.current_frame().move_selection(ez_3D_float(ail.move_points_per_key, 0, 0));
                                ui_message = "+ X move";
                                break;                    
                            //------------------------------------------------
                            case 'X':
                                ail.current_frame().move_selection(ez_3D_float(-ail.move_points_per_key, 0, 0));
                                ui_message = "- X move";
                                break;                    
                            //------------------------------------------------
                            case 'y':
                                ail.current_frame().move_selection(ez_3D_float(0, ail.move_points_per_key, 0));
                                ui_message = "+ Y move";
                                break;                    
                            //------------------------------------------------
                            case 'Y':
                                ail.current_frame().move_selection(ez_3D_float(0, -ail.move_points_per_key, 0));
                                ui_message = "- Y move";
                                break;                    
                            //------------------------------------------------
                            case 'z':
                                ail.current_frame().move_selection(ez_3D_float(0, 0, ail.move_points_per_key));
                                ui_message = "+ Z move";
                                break;                    
                            //------------------------------------------------
                            case 'Z':
                                ail.current_frame().move_selection(ez_3D_float(0, 0,-ail.move_points_per_key));
                                ui_message = "- Z move";
                                break;
                            //------------------------------------------------
                            case 'f':
                                ail.current_frame().rotate_selection(ez_3D_float(0, 0, ail.rotate_radians_per_key));
                                ui_message = "+ rotate Z";
                                break;                    
                            //------------------------------------------------
                            case 'F':
                                ail.current_frame().rotate_selection(ez_3D_float(0, 0, -ail.rotate_radians_per_key));
                                ui_message = "- rotate Z";
                                break;                    
                            //------------------------------------------------
                            case 'g':
                                ail.current_frame().rotate_selection(ez_3D_float(0, ail.rotate_radians_per_key, 0));
                                ui_message = "+ rotate y";
                                break;                    
                            //------------------------------------------------
                            case 'G':
                                ail.current_frame().rotate_selection(ez_3D_float(0,-ail.rotate_radians_per_key, 0));
                                ui_message = "- rotate y";
                                break;                    
                            //------------------------------------------------
                            case 'r':
                                ail.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, 0, ail.rotate_radians_per_key));
                                ui_message = "+ rotate Z on a point";
                                break;                    
                            //------------------------------------------------
                            case 'R':
                                ail.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, 0, -ail.rotate_radians_per_key));
                                ui_message = "- rotate Z on a point";
                                break;                    
                            //------------------------------------------------
                            case 't':
                                ail.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, ail.rotate_radians_per_key, 0));
                                ui_message = "+ rotate Y on a point";
                                break;                    
                            //------------------------------------------------
                            case 'T':
                                ail.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, -ail.rotate_radians_per_key, 0));
                                ui_message = "- rotate Y on a point";
                                break;                    
                            //------------------------------------------------
                            case '9':
                                ail.current_frame().scale_selection(ez_3D_float(1.01, 1, 1));
                                ui_message = "+ scale X";
                                break;                    
                            //------------------------------------------------
                            case '(':
                                ail.current_frame().scale_selection(ez_3D_float(0.99, 1, 1));
                                ui_message = "- scale X";
                                break;                    
                            //------------------------------------------------
                            case '0':
                                ail.current_frame().scale_selection(ez_3D_float(1, 1.01, 1));
                                ui_message = "+ scale Y";
                                break;                    
                            //------------------------------------------------
                            case ')':
                                ail.current_frame().scale_selection(ez_3D_float(1, 0.99, 1));
                                ui_message = "- scale Y";
                                break;                    
                            //------------------------------------------------
                            case '-':
                                ail.current_frame().scale_selection(ez_3D_float(1, 1, 1.01));
                                ui_message = "+ scale Z";
                                break;                    
                            //------------------------------------------------
                            case '_':
                                ail.current_frame().scale_selection(ez_3D_float(1, 1, 0.99));
                                ui_message = "- scale Z";
                                break;                    
                            //------------------------------------------------
                            case '=':
                                ail.current_frame().scale_selection(ez_3D_float(1.01, 1.01, 1.01));
                                ui_message = "+ scale XYZ";
                                break;                    
                            //------------------------------------------------
                            case '+':
                                ail.current_frame().scale_selection(ez_3D_float(0.99, 0.99, 0.99));
                                ui_message = "- scale XYZ";
                                break;                    
                            //------------------------------------------------
                            case 'v':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1.01, 1, 1));
                                ui_message = "+ scale X by a point";
                                break;                    
                            //------------------------------------------------
                            case 'V':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(0.99, 1, 1));
                                ui_message = "- scale X by a point";
                                break;                    
                            //------------------------------------------------
                            case 'b':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 1.01, 1));
                                ui_message = "+ scale Y by a point";
                                break;                    
                            //------------------------------------------------
                            case 'B':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 0.99, 1));
                                ui_message = "- scale Y by a point";
                                break;                    
                            //------------------------------------------------
                            case 'n':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 1, 1.01));
                                ui_message = "+ size Z by a point";
                                break;                    
                            //------------------------------------------------
                            case 'N':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 1, 0.99));
                                ui_message = "- scale Z by a point";
                                break;                    
                            //------------------------------------------------
                            case 'm':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1.01, 1.01, 1.01));
                                ui_message = "+ scale XYZ by a point";
                                break;                    
                            //------------------------------------------------
                            case 'M':
                                ail.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(0.99, 0.99, 0.99));
                                ui_message = "- scale XYZ by a point";
                                break;                    
                            //------------------------------------------------
                            case 'a':
                                ail.arbitrary_point.x += ail.move_points_per_key;
                                ui_message = "+ X a point";
                                break;                    
                            //------------------------------------------------
                            case 'A':
                                ail.arbitrary_point.x -= ail.move_points_per_key;
                                ui_message = "- X a point";
                                break;                    
                            //------------------------------------------------
                            case 's':
                                ail.arbitrary_point.y += ail.move_points_per_key;
                                ui_message = "+ Y a point";
                                break;                    
                            //------------------------------------------------
                            case 'S':
                                ail.arbitrary_point.y -= ail.move_points_per_key;
                                ui_message = "- Y a point";
                                break;                    
                            //------------------------------------------------
                            case 'd':
                                ail.arbitrary_point.z += ail.move_points_per_key;
                                ui_message = "+ Z a point";
                                break;                    
                            //------------------------------------------------
                            case 'D':
                                ail.arbitrary_point.z -= ail.move_points_per_key;
                                ui_message = "- Z a point";
                                break;                    
                            //------------------------------------------------
                            case ';': // toggle bitmap background
                                ail.show_background_bitmap = !ail.show_background_bitmap;
                                break;                    
                            //------------------------------------------------
                              case SDLK_RETURN:
                                show_help = !show_help;
                            //------------------------------------------------
                        } // end switch(event.key.keysym.unicode)
                        ail.display(ui_message);
                        if(show_help)
                            ail.display_move_scale_rotate_menu();
                        if(show_frame_stats)
                            ail.current_frame().display_stats();
                        if(show_frame_set_stats)
                            ail.display_stats();
                    } // end else if(event.type == SDL_KEYDOWN)
                } // end while(in_sub_menu)
                show_help = show_help_was;
                break;
            }   // end local scope
            //----------------------------------------------------------------






            //----------------------------------------------------------------
            case 'b': // LaserBoy wave utilities
            case 'B':
                in_sub_menu = true;
                ui_message = "wave utilitites";
                ail.display(ui_message);
                ail.display_wave_utilities_menu();                
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "exit wave utilities";
                                break;
                            //------------------------------------------------
                            case '1': // LaserBoy wave utilities settings
                            {
                                bool in_sub_sub_menu = true;
                                ail.display("wave show settings");
                                ail.display_wave_show_settings_menu();
                                while(in_sub_sub_menu)
                                {
                                    SDL_WaitEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                    {
                                        switch(event.key.keysym.unicode)
                                        {
                                            //--------------------------------
                                            case SDLK_ESCAPE:
                                                ui_message = "action canceled";
                                                in_sub_sub_menu = false;
                                                break;
                                            //--------------------------------
                                            case '1':
                                                ail.wave_show_points = !ail.wave_show_points;
                                                if(ail.wave_show_points)
                                                    ui_message = "show vertices";
                                                else
                                                    ui_message = "hide vertices";
                                                break;
                                            //--------------------------------
                                            case '2':
                                                ail.wave_show_blanks = !ail.wave_show_blanks;
                                                if(ail.wave_show_blanks)
                                                    ui_message = "show blanking";
                                                else
                                                    ui_message = "hide blanking";
                                                break;
                                            //--------------------------------
                                            case '3':
                                                ail.wave_real_time = !ail.wave_real_time;
                                                if(ail.wave_real_time)
                                                    ui_message = "real time";
                                                else
                                                    ui_message = "as fast as possible";
                                                break;
                                            //--------------------------------
                                            case '4':
                                                ail.wave_show_inverted = !ail.wave_show_inverted;
                                                if(ail.wave_show_inverted)
                                                    ui_message = "inverted";
                                                else
                                                    ui_message = "non-inverted";
                                                break;
                                            //--------------------------------
                                            case '5':
                                                ail.wave_clean_screen = !ail.wave_clean_screen;
                                                if(ail.wave_clean_screen)
                                                    ui_message = "clean screen";
                                                else
                                                    ui_message = "no clean screen";
                                                break;
                                            //--------------------------------
                                            case '6':
                                                ail.wave_loop = !ail.wave_loop;
                                                if(ail.wave_loop)
                                                    ui_message = "loop";
                                                else
                                                    ui_message = "no loop";
                                                break;
                                            //--------------------------------
                                        }
                                    }
                                    ail.display(ui_message);
                                    ail.display_wave_show_settings_menu();                                    
                                }
                                break;
                            }                                
                            //------------------------------------------------
                            case '2':
                                ail.display_file_list(EZ_UAIL_WAVE_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter wave file name");
                                if(ail.process_exception)
                                    break;                                
                                ail.display("wave stats");
                                ail.display_wave_stats(file_name);
                                while(true)
                                {
                                    SDL_WaitEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                        break;
                                }
                                break;
                            //------------------------------------------------
                            case '3': // show LaserBoy wave on screen
                            {
                                bool playing = true;
                                ail.display_file_list(EZ_UAIL_WAVE_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter wave file name");
                                if(ail.process_exception)
                                    break;
                                do
                                {                                    
                                    ail.display_wave(file_name);
                                    SDL_PollEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                        break;
                                } while(ail.wave_loop && playing);
                                
                                while(true)
                                {
                                    SDL_WaitEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                        break;
                                }
                                break;
                            }
                            //------------------------------------------------
                            case '4': // set LaserBoy wave offsets
                            {                     
                                ail.display("sample offsets");                                
                                ail.display_wave_offsets();
                                temp_int = ail.display_prompt_and_echo_int("enter wave sample offset for both X and Y");
                                if(ail.process_exception)
                                    break;
                                ail.wave_offsets[0] = temp_int;                                
                                ail.wave_offsets[1] = temp_int;
                                ail.display("sample offsets");                                
                                ail.display_wave_offsets();
                                if(ail.process_exception)
                                    break;
                                ail.display("sample offsets");                                
                                ail.display_wave_offsets();
                                temp_int = ail.display_prompt_and_echo_int("enter wave sample offset for red");
                                if(ail.process_exception)
                                    break;
                                ail.wave_offsets[2] = temp_int;
                                ail.display("sample offsets");                                
                                ail.display_wave_offsets();
                                temp_int = ail.display_prompt_and_echo_int("enter wave sample offset for green");
                                if(ail.process_exception)
                                    break;
                                ail.wave_offsets[3] = temp_int;
                                ail.display("sample offsets");                                
                                ail.display_wave_offsets();
                                temp_int = ail.display_prompt_and_echo_int("enter wave sample offset for blue");
                                if(ail.process_exception)
                                    break;
                                ail.wave_offsets[4] = temp_int;
                                ail.display("sample offsets");                                
                                ail.display_wave_offsets();
                                temp_int = ail.display_prompt_and_echo_int("enter wave sample offset for monochrome");
                                if(ail.process_exception)
                                    break;
                                ail.wave_offsets[5] = temp_int;
                                ail.display("sample offsets");                                
                                ail.display_wave_offsets();
                                while(true)
                                {
                                    SDL_WaitEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                        break;
                                }
                                break;
                            }
                            //------------------------------------------------
                            case '5': // apply LaserBoy wave sample offsets                            
                                ail.display_file_list(EZ_UAIL_WAVE_SHARE);                                
                                file_name = ail.display_prompt_and_echo_name("enter wave file name");
                                if(ail.process_exception)
                                    break;
                                ail.apply_wave_offsets(file_name);
                                if(!ail.process_exception)
                                {
                                    ail.display("wave stats");
                                    ail.display_wave_stats(file_name);
                                }
                                while(true)
                                {
                                    SDL_WaitEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                        break;
                                }
                                break;
                            //------------------------------------------------
                            case '6': // overwrite LaserBoy wave sample offsets
                                ail.display_file_list(EZ_UAIL_WAVE_SHARE);                                
                                file_name = ail.display_prompt_and_echo_name("enter wave file name");
                                if(ail.process_exception)
                                    break;
                                ail.overwirte_wave_offsets(file_name);
                                if(!ail.process_exception)
                                {
                                    ail.display("wave stats");
                                    ail.display_wave_stats(file_name);
                                }
                                while(true)
                                {
                                    SDL_WaitEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                        break;
                                }
                                break;
                            //------------------------------------------------
                            case '7': // invert LaserBoy wave polarity
                                ail.display_file_list(EZ_UAIL_WAVE_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter wave file name");
                                if(ail.process_exception)
                                    break;
                                ail.invert_wave(file_name);
                                if(!ail.process_exception)
                                {
                                    ail.display("wave stats");
                                    ail.display_wave_stats(file_name);
                                }
                                while(true)
                                {
                                    SDL_WaitEvent(&event);
                                    if(event.type == SDL_KEYDOWN)
                                        break;
                                }
                                break;
                            //------------------------------------------------
                            case '8': // split LaserBoy wave
                                ail.display_file_list(EZ_UAIL_WAVE_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter wave file name");
                                if(ail.process_exception)
                                    break;
                                ail.display("splitting wave");
                                ail.display_please_wait();
                                ail.split_wave_xyrgbi(file_name);
                                ail.display("split wave");
                                break;
                            //------------------------------------------------
                            case '9': // join waves to LaserBoy wave
                            {
                                string file_xy ,
                                       file_r  ,
                                       file_g  ,
                                       file_b  ,
                                       file_i  ,
                                       file_out;
                                ail.display_file_list(EZ_UAIL_SPLIT_WAVE_SHARE);
                                file_xy = ail.display_prompt_and_echo_name("enter XY wave");
                                if(ail.process_exception)
                                    break;
                                ail.display("opening wave");
                                ail.display_file_list(EZ_UAIL_SPLIT_WAVE_SHARE);
                                file_r = ail.display_prompt_and_echo_name("enter red wave");
                                if(ail.process_exception)
                                    break;
                                ail.display("opening wave");
                                ail.display_file_list(EZ_UAIL_SPLIT_WAVE_SHARE);
                                file_g = ail.display_prompt_and_echo_name("enter green wave");
                                if(ail.process_exception)
                                    break;
                                ail.display("opening wave");
                                ail.display_file_list(EZ_UAIL_SPLIT_WAVE_SHARE);
                                file_b = ail.display_prompt_and_echo_name("enter blue wave");
                                if(ail.process_exception)
                                    break;
                                ail.display("opening wave");
                                ail.display_file_list(EZ_UAIL_SPLIT_WAVE_SHARE);
                                file_i = ail.display_prompt_and_echo_name("enter intensity wave");
                                if(ail.process_exception)
                                    break;
                                ail.display("new wave name");
                                ail.display_file_list(EZ_UAIL_WAVE_SHARE);
                                file_out = ail.display_prompt_and_echo_name("enter new name for wave out");
                                if(ail.process_exception)
                                    break;
                                ail.display("joining waves");
                                ail.display_please_wait();
                                ail.join_xyrgbi_waves(file_xy, file_r, file_g, file_b, file_i, file_out);
                                ail.display("joined waves");
                                break;
                            }
                            //------------------------------------------------
                        }
                        ail.display(ui_message);
                        ail.display_wave_utilities_menu();
                    }
                }
                break;
            //----------------------------------------------------------------                                        




            //----------------------------------------------------------------
            case 'i': // open data from file
            case 'I':
                ui_message = "input file";
                ail.display(ui_message);
                ail.display_in_file_type_menu();
                do
                {
                    SDL_WaitEvent(&event);
                } while(event.type != SDL_KEYDOWN);
                switch(event.key.keysym.unicode)
                {
                    //--------------------------------------------------------
                    case SDLK_ESCAPE:
                        in_sub_menu = false;
                        ui_message = "action canceled";
                        break;
                    //--------------------------------------------------------
                    case '1': // open ILDA
                        ail.display();
                        ail.display_file_list(EZ_UAIL_FRAME_SHARE);
                        file_name = ail.display_prompt_and_echo_name("enter ILDA file name");
                        if(ail.process_exception)
                            break;                            
                        ail.display();
                        ail.display_in_ild_action_menu();            
                        do
                        {
                            SDL_WaitEvent(&event);
                        } while(event.type != SDL_KEYDOWN);
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                ui_message = "action canceled";
                                break;
                            //------------------------------------------------
                            case '1': // replace
                                ail.display("opening ILDA");
                                ail.display_please_wait();
                                ail.from_file(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "replace";
                                break;
                            //------------------------------------------------
                            case '2': // prepend
                                ail.display("opening ILDA");
                                ail.display_please_wait();
                                {
                                    ez_uail_frame_set prepend;
                                    prepend.from_file(file_name);
                                    if(ail.process_exception)
                                    {
                                        while(true)
                                        {
                                            SDL_WaitEvent(&event);
                                            if(event.type == SDL_KEYDOWN)
                                                break;
                                        }
                                        break;                            
                                    }
                                    prepend += ail;
                                    ail = prepend;
                                }                          
                                ui_message = "prepend";
                                break;
                            //------------------------------------------------
                            case '3': // insert
                                ail.display("opening ILDA");
                                ail.display_please_wait();
                                {
                                    int                i;
                                    ez_uail_frame_set  insert,
                                                       result;
                                    insert.from_file(file_name);                                                    
                                    if(ail.process_exception)
                                    {
                                        while(true)
                                        {
                                            SDL_WaitEvent(&event);
                                            if(event.type == SDL_KEYDOWN)
                                                break;
                                        }
                                        break;                            
                                    }
                                    for(i = 0; i <= ail.frame_index; i++)
                                        result += ail[i];
                                    for(i = 0; i < insert.number_of_frames(); i++)
                                        result += insert[i];
                                    for(i = ail.frame_index + 1; i < ail.number_of_frames(); i++)
                                        result += ail[i];
                                    ail = result;
                                }
                                ail.from_file(file_name, true);
                                ui_message = "insert";
                                break;
                            //------------------------------------------------
                            case '4': // append
                                ail.display("opening ILDA");
                                ail.display_please_wait();
                                ail.from_file(file_name, true);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "append";
                                break;
                            //------------------------------------------------
                            case '5': // superimpose
                                ail.display("opening ILDA");
                                ail.display_please_wait();
                                ail.superimpose_from_file(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "superimpose";
                                break;
                            //------------------------------------------------
                        }
                        break;
                    //--------------------------------------------------------
                    case '2': // open DXF
                        ail.display();
                        ail.display_in_dxf_action_menu();
                        do
                        {
                            SDL_WaitEvent(&event);
                        } while(event.type != SDL_KEYDOWN);
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                ui_message = "action canceled";
                                break;
                            //------------------------------------------------
                            case '1': // open dxf
                                ail.display_file_list(EZ_UAIL_DXF_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter DXF file name");
                                if(ail.process_exception)
                                    break;
                                ail.display("opening DXF");
                                ail.display_please_wait();
                                ail.from_dxf_file(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "open dxf";
                                break;
                            //------------------------------------------------
                            case '2': // append dxf
                                ail.display_file_list(EZ_UAIL_DXF_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter DXF file name");            
                                if(ail.process_exception)
                                    break;
                                ail.display("opening DXF");
                                ail.display_please_wait();
                                ail.from_dxf_file(file_name, true);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "append dxf file";
                                break;
                            //------------------------------------------------
                            case '3': // open dxf directory
                                ail.display_directory_list(EZ_UAIL_DXF_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter DXF directory name");
                                if(ail.process_exception)
                                    break;
                                ail.display("opening DXF");
                                ail.display_please_wait();
                                ail.from_dxf_directory(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "open dxf directory";
                                break;
                            //------------------------------------------------
                            case '4': // append dxf directory
                                ail.display_directory_list(EZ_UAIL_DXF_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter DXF directory name");
                                if(ail.process_exception)
                                    break;
                                ail.display("opening DXF");
                                ail.display_please_wait();
                                ail.from_dxf_directory(file_name, true);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "open dxf directory";
                                break;
                            //------------------------------------------------
                        }
                        break;
                    //--------------------------------------------------------
                    case '3': // open LaserBoy wave                                
                        ail.display();
                        ail.display_file_list(EZ_UAIL_WAVE_SHARE);                            
                        file_name = ail.display_prompt_and_echo_name("enter LaserBoy wave file name");
                        if(ail.process_exception)
                            break;
                        ail.display();
                        ail.display_in_wave_action_menu();
                        do
                        {
                            SDL_WaitEvent(&event);
                        } while(event.type != SDL_KEYDOWN);
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                ui_message = "action canceled";
                                break;
                            //------------------------------------------------
                            case '1': // open wave
                                ail.display("opening wave");
                                ail.display_please_wait();
                                ail.from_wave_file(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "wave open";
                                break;
                            //------------------------------------------------
                            case '2': // append wave
                                ail.display("opening wave");
                                ail.display_please_wait();
                                ail.from_wave_file(file_name, true);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "wave append";
                                break;
                            //------------------------------------------------
                        }
                        break;
                    //--------------------------------------------------------
                    case '4': // import ADAT wave set
                    {
                        string file_a,
                               file_b,
                               file_c,
                               file_d;
                        ail.display_file_list(EZ_UAIL_ADAT_SHARE);
                        file_a = ail.display_prompt_and_echo_name("enter ADAT XY wave");
                        if(ail.process_exception)
                            break;
                        ail.display("opening wave");
                        file_b = ail.display_prompt_and_echo_name("enter ADAT red wave");
                        if(ail.process_exception)
                            break;
                        ail.display("opening wave");
                        file_c = ail.display_prompt_and_echo_name("enter ADAT green wave");
                        if(ail.process_exception)
                            break;
                        ail.display("opening wave");
                        file_d = ail.display_prompt_and_echo_name("enter ADAT blue wave");
                        if(ail.process_exception)
                            break;
                        ail.display("ADAT import");
                        ail.display_please_wait();
                        ail.from_ADAT_wave_files(file_a, file_b, file_c, file_d);
                        if(ail.process_exception)
                        {
                            while(true)
                            {
                                SDL_WaitEvent(&event);
                                if(event.type == SDL_KEYDOWN)
                                    break;
                            }
                            break;                            
                        }
                        break;
                    }
                    //--------------------------------------------------------
                    case '5': // open bitmap
                        ail.display();
                        ail.display_in_bmp_action_menu();
                        do
                        {
                            SDL_WaitEvent(&event);
                        } while(event.type != SDL_KEYDOWN);
                        switch(event.key.keysym.unicode)
                        {
                            //------------------------------------------------
                            case SDLK_ESCAPE:
                                ui_message = "action canceled";
                                break;
                            //------------------------------------------------
                            case '1': //
                                ail.display_file_list(EZ_UAIL_BMP_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter bitmap file name");
                                if(ail.process_exception)
                                    break;
                                ail.display("loading bmp");
                                ail.display_please_wait();
                                ail.current_frame().color_over_ezfb_bitmap(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "color over bmp frame";
                                break;
                            //------------------------------------------------
                            case '2': //
                                ail.display_file_list(EZ_UAIL_BMP_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter bitmap file name");
                                if(ail.process_exception)
                                    break;
                                ail.display("loading bmp");
                                ail.display_please_wait();
                                ail.color_over_ezfb_bitmap(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "color over bmp all";
                                break;
                            //------------------------------------------------
                            case '3': //
                                ail.display_file_list(EZ_UAIL_BMP_SHARE);
                                file_name = ail.display_prompt_and_echo_name("enter bitmap file name");
                                if(ail.process_exception)
                                    break;
                                ail.display("loading bmp");
                                ail.display_please_wait();
                                ail.load_background_bitmap(file_name);
                                if(ail.process_exception)
                                {
                                    while(true)
                                    {
                                        SDL_WaitEvent(&event);
                                        if(event.type == SDL_KEYDOWN)
                                            break;
                                    }
                                    break;                            
                                }
                                ui_message = "background bmp";
                                break;
                            //------------------------------------------------
                            case '4': //
                                ail.clear_background_bitmap();
                                ui_message = "clear background";
                                break;
                        //----------------------------------------------------
                        }
                        break;
                    //--------------------------------------------------------
                    case '6': // ascii palette
                        ail.display_file_list(EZ_UAIL_PALETTE_SHARE);
                        file_name = ail.display_prompt_and_echo_name("enter Palette file name");
                        if(ail.process_exception)
                            break;
                        ail.display("opening palette");
                        ail.display_please_wait();
                        ail.add_palette_from_file(file_name);
                        if(ail.process_exception)
                        {
                            while(true)
                            {
                                SDL_WaitEvent(&event);
                                if(event.type == SDL_KEYDOWN)
                                    break;
                            }
                            break;                            
                        }
                        ui_message = "open palette";
                        break;
                    //--------------------------------------------------------
                } // end switch(event.key.keysym.unicode)
                break;
                //------------------------------------------------------------







            //----------------------------------------------------------------
            case 'o': // save contents of memory to various file types
            case 'O':
                ui_message = "save as file"; 
                ail.display(ui_message);
                ail.display_out_file_type_menu();
                do
                {
                    SDL_WaitEvent(&event);
                } while(event.type != SDL_KEYDOWN);
                switch(event.key.keysym.unicode)
                {
                    //--------------------------------------------------------
                    case SDLK_ESCAPE:
                        in_sub_menu = false;
                        ui_message = "action canceled";
                        break;
                    //--------------------------------------------------------
                    case '1':   
                        file_name = ail.display_prompt_and_echo_name("enter new ILDA file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving ILDA");
                        ail.display_please_wait();
                        ail.current_frame().save_as_uail(file_name);
                        ui_message = "ILDA saved";
                        break;
                    //--------------------------------------------------------
                    case '2':   
                        file_name = ail.display_prompt_and_echo_name("enter new ILDA file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving ILDA");
                        ail.display_please_wait();
                        ail.selected_frames().save_as_uail(file_name);
                        ui_message = "ILDA saved";
                        break;
                    //--------------------------------------------------------
                    case '3':   
                        file_name = ail.display_prompt_and_echo_name("enter new ILDA file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving ILDA");
                        ail.display_please_wait();
                        ail.save_as_uail(file_name);
                        ui_message = "ILDA saved";
                        break;
                    //--------------------------------------------------------
                    case '4':   
                        ail.display_frame_effects();
                        file_name = ail.display_prompt_and_echo_name("enter frame effect name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving ILDA");
                        ail.display_please_wait();
                        ail.call_frame_effect(file_name);
                        ui_message = "ILDA saved";
                        break;
                    //--------------------------------------------------------
                    case '5':   
                        ail.display_frame_set_effects();
                        file_name = ail.display_prompt_and_echo_name("enter frame set effect name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving ILDA");
                        ail.display_please_wait();
                        ail.call_frame_set_effect(file_name);
                        ui_message = "ILDA saved";
                        break;
                    //--------------------------------------------------------
                    case '6':   
                        file_name = ail.display_prompt_and_echo_name("enter new DXF file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving DXF");
                        ail.display_please_wait();
                        ail.current_frame().save_as_dxf(file_name);
                        ui_message = "DXF saved";
                        break;
                    //--------------------------------------------------------
                    case '7':
                        file_name = ail.display_prompt_and_echo_name("enter new DXF directory name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving DXF");
                        ail.display_please_wait();
                        ail.selected_frames().save_as_dxf_directory(file_name);
                        ui_message = "DXF directory saved";
                        break;
                    //--------------------------------------------------------
                    case '8':
                        file_name = ail.display_prompt_and_echo_name("enter new DXF directory name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving DXF");
                        ail.display_please_wait();
                        ail.save_as_dxf_directory(file_name);
                        ui_message = "DXF directory saved";
                        break;
                    //--------------------------------------------------------
                    case '9':
                        file_name = ail.display_prompt_and_echo_name("enter new wave file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving wave");
                        ail.display_please_wait();
                        ail.save_as_wave(file_name);
                        ui_message = "wave saved";
                        break;
                    //--------------------------------------------------------
                    case '0':
                        file_name = ail.display_prompt_and_echo_name("enter new wave file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving wave");
                        ail.display_please_wait();
                        ail.current_frame().save_as_wave(file_name);
                        ui_message = "wave saved";
                        break;
                    //--------------------------------------------------------
                    case 'a':   
                        file_name = ail.display_prompt_and_echo_name("enter new C++ file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving C++");
                        ail.display_please_wait();
                        ail.current_frame().save_as_code(file_name);
                        ui_message = "C++ saved";
                        break;
                    //--------------------------------------------------------
                    case 'b':   
                        file_name = ail.display_prompt_and_echo_name("enter new C++ file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving C++");
                        ail.display_please_wait();
                        ail.selected_frames().save_as_code(file_name);
                        ui_message = "C++ saved";
                        break;
                    //--------------------------------------------------------
                    case 'c':
                        file_name = ail.display_prompt_and_echo_name("enter new C++ file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving C++");
                        ail.display_please_wait();
                        ail.save_as_code(file_name);
                        ui_message = "C++ saved";
                        break;
                    //--------------------------------------------------------
                    case 'd':
                        file_name = ail.display_prompt_and_echo_name("enter new bitmap file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving bitmap");
                        ail.display_please_wait();
                        ail.current_frame().save_as_bmp(file_name);
                        ui_message = "bitmap saved";
                        break;
                    //--------------------------------------------------------
                    case 'e':
                        file_name = ail.display_prompt_and_echo_name("enter new bitmap directory file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving bitmap directory");
                        ail.display_please_wait();
                        ail.selected_frames().save_as_bmp_directory(file_name);
                        ui_message = "bitmap directory saved";
                        break;
                    //--------------------------------------------------------
                    case 'f':
                        file_name = ail.display_prompt_and_echo_name("enter new bitmap directory file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving bitmap directory");
                        ail.display_please_wait();
                        ail.save_as_bmp_directory(file_name);
                        ui_message = "bitmap directory saved";
                        break;
                    //--------------------------------------------------------
                    case 'g':
                        file_name = ail.display_prompt_and_echo_name("enter new palette file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving ascii palette");
                        ail.display_please_wait();
                        ail.palette_picker(ail.current_frame().palette_index).save_as_ascii(file_name);
                        ui_message = "ascii palette saved";
                        break;
                    //--------------------------------------------------------
                    case 'h':
                        file_name = ail.display_prompt_and_echo_name("enter new palette file name");
                        if(ail.process_exception)
                            break;
                        ail.display("saving ascii palette");
                        ail.display_please_wait();
                        ail.palette_picker(ail.target_palette_index).save_as_ascii(file_name);
                        ui_message = "ascii palette saved";
                        break;
                    //--------------------------------------------------------
                    case 'i':
                        ail.save_cfg_file();
                        ui_message = "LaserBoy.wtf saved";
                        break;
                    //--------------------------------------------------------
                } // end switch(event.key.keysym.unicode)
                break;
            //----------------------------------------------------------------
        } // end switch(event.key.keysym.unicode)
        ail.display(ui_message);
        if(show_help)
            ail.display_program_menu();
        if(show_frame_stats)
            ail.current_frame().display_stats();
        if(show_frame_set_stats)
            ail.display_stats();
        if(teminate_requested)
        {
            bg_was = ail.bg_color;
            ail.bg_color = (u_int)0x00ff0000;
            ail.recolor_background();
            ail.display("terminating");
            if(ail.display_prompt_and_echo_bool("Do you really want to terminate laserboy? y|Y"))
                running = false;
            else
            {
                ail.bg_color = bg_was;
                ail.recolor_background();
                teminate_requested = false;
                ui_message = "terminate canceled";                
                ail.display(ui_message);
                if(show_help)
                    ail.display_program_menu();
            }
        }
    } // end while(running)
    ail.bg_color = (u_int)0x000000ff;
    ail.recolor_background();
    ail.display("terminating");
    if(!ail.display_prompt_and_echo_bool("Exit without saving OUT.ild? y|Y"))
    {   
        ail.display("saving OUT.ild");
        if(ail.auto_optimize)
            ail.minimize_points();
        ail.save_as_uail("OUT.ild");
    }
    else
    //------------------------------------------------------------------------
    ail.display("LaserBoy terminated");
    //------------------------------------------------------------------------
    SDL_Delay(500);
    return(0);
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
